"""
compute rolling correlations between crypto and asset
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import qis as qis
import yfinance as yf

# define asset and cryptocurrency
ASSET = 'QQQ'
CRYPTO = 'BTC-USD'
tickers = [ASSET, CRYPTO]
# fetch yahoo data
prices = yf.download(tickers, start=None, end=None)['Close'][tickers]
# resample to business days
prices = prices.asfreq('B', method='ffill').dropna()
# % returns
returns = prices.pct_change()
# returns = np.log(prices).diff()

# compute rolling correlations
corr_3m = returns[CRYPTO].rolling(63).corr(returns[ASSET]).rename('3m')
corr_1y = returns[CRYPTO].rolling(252).corr(returns[ASSET]).rename('1y')
corr_2y = returns[CRYPTO].rolling(2*252).corr(returns[ASSET]).rename('2y')
corrs = pd.concat([corr_3m, corr_1y, corr_2y], axis=1).dropna()

# select period
time_period = qis.TimePeriod('01Jan2016', None)
corrs = time_period.locate(corrs)
# qis.save_df_to_excel(data=corrs, file_name='btc_spy_corr')

with sns.axes_style("darkgrid"):
    fig, ax = plt.subplots(1, 1, figsize=(15, 8), tight_layout=True)

qis.plot_time_series(df=corrs,
                     trend_line=qis.TrendLine.ZERO_SHADOWS,
                     legend_stats=qis.LegendStats.AVG_LAST_SCORE,
                     var_format='{:.0%}',
                     fontsize=14,
                     title=f"Rolling correlation of daily returns between {CRYPTO} and {ASSET} as function of rolling window",
                     ax=ax)
# qis.save_fig(fig=fig, file_name='btc_all_corr')

plt.show()





from qis.perfstats.config import (
    FULL_TABLE_COLUMNS,
    PerfParams,
    PerfStat,
    RA_TABLE_COLUMNS,
    RA_TABLE_COMPACT_COLUMNS,
    RegimeData,
    RegimeType,
    ReturnTypes,
    SD_PERF_COLUMNS,
    TRE_TABLE_COLUMNS
)

from qis.perfstats.cond_regression import estimate_cond_regression, get_regime_regression_params

from qis.perfstats.desc_table import DescTableType, compute_desc_table

from qis.perfstats.perf_stats import (
    STANDARD_TABLE_COLUMNS,
    LN_TABLE_COLUMNS,
    LN_BENCHMARK_TABLE_COLUMNS,
    LN_BENCHMARK_TABLE_COLUMNS_SHORT,
    EXTENDED_TABLE_COLUMNS,
    COMPACT_TABLE_COLUMNS,
    BENCHMARK_TABLE_COLUMNS,
    BENCHMARK_TABLE_COLUMNS2,
    compute_avg_max_dd,
    compute_desc_freq_table,
    compute_rolling_drawdowns,
    compute_rolling_drawdown_time_under_water,
    compute_info_ratio_table,
    compute_max_dd,
    compute_performance_table,
    compute_ra_perf_table,
    compute_ra_perf_table_with_benchmark,
    compute_risk_table,
    compute_te_ir_errors,
    compute_drawdowns_stats_table
)

from qis.perfstats.regime_classifier import (
    BenchmarkReturnsQuantileRegimeSpecs,
    BenchmarkReturnsQuantilesRegime,
    BenchmarkVolsQuantilesRegime,
    RegimeClassifier,
    VolQuantileRegimeSpecs,
    compute_bnb_regimes_pa_perf_table,
    compute_mean_freq_regimes,
    compute_regime_avg,
    compute_regimes_pa_perf_table_from_sampled_returns
)

from qis.perfstats.returns import (
    adjust_navs_to_portfolio_pa,
    compute_excess_returns,
    compute_grouped_nav,
    compute_net_return,
    compute_num_years,
    compute_pa_excess_returns,
    compute_pa_return,
    compute_returns_dict,
    compute_sampled_vols,
    compute_total_return,
    estimate_vol,
    get_excess_returns_nav,
    get_net_navs,
    log_returns_to_nav,
    portfolio_navs_to_additive,
    portfolio_returns_to_nav,
    prices_at_freq,
    returns_to_nav,
    to_portfolio_returns,
    long_short_to_relative_nav,
    to_returns,
    prices_to_scaled_nav,
    to_total_returns,
    to_zero_first_nonnan_returns,
    df_price_ffill_between_nans
)

from qis.perfstats.timeseries_bfill import (
    interpolate_infrequent_returns,
    append_time_series,
    bfill_timeseries,
    df_fill_first_nan_by_cross_median,
    df_price_fill_first_nan_by_cross_median,
    replace_nan_by_median,
    df_ffill_negatives
)

from qis.perfstats.fx_ops import (get_aligned_fx_spots,
                                  compute_futures_fx_adjusted_returns)



"""
pieplot
"""
# packages
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from typing import Optional, List
from enum import Enum

# qis
import qis.plots.utils as put


def plot_pie(df: [pd.Series, pd.DataFrame],
             y_column: str = None,
             ylabel: str = '',
             title: str = None,
             colors: List[str] = None,
             legend_loc: Optional[str] = None,
             autopct: Optional[str] = '%.0f%%',
             ax: plt.Subplot = None,
             **kwargs
             ) -> Optional[plt.Figure]:

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if y_column is None and isinstance(df, pd.DataFrame):
        y_column = df.columns[0]
    if colors is None:
        colors = put.get_cmap_colors(n=len(df.index), **kwargs)

    df.plot.pie(y=y_column, autopct=autopct, colors=colors, ax=ax)

    if legend_loc is None:
        ax.legend().set_visible(False)

    if title is not None:
        put.set_title(ax=ax, title=title, **kwargs)

    ax.set_ylabel(ylabel)

    return fig


class UnitTests(Enum):
    PORTFOLIO = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.PORTFOLIO:

        df = pd.DataFrame({'Conservative': [0.5, 0.25, 0.25],
                           'Balanced': [0.30, 0.30, 0.40],
                           'Growth': [0.10, 0.40, 0.50]},
                          index=['Stables', 'Market-neutral', 'Crypto-Beta'])
        print(df)
        kwargs = dict(fontsize=8, linewidth=0.5, weight='normal', markersize=1)

        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(8, 6), tight_layout=True)
            plot_pie(df=df,
                     ax=ax,
                     **kwargs)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PORTFOLIO

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List
from enum import Enum

# qis
import yfinance as yf
import qis

SEED = 1


def plot_bootsrap_paths(prices: pd.Series,
                        num_samples: int = 10,
                        block_size: int = 20,
                        nlags: int = 20,
                        ) -> List[plt.Figure]:
    """
    create analysis figure
    """
    figs = []
    bootstrap_prices = qis.bootstrap_price_data(prices=prices,
                                                bootsrap_type=qis.BootsrapType.STATIONARY,
                                                bootsrap_output=qis.BootsrapOutput.SERIES_TO_DF,
                                                num_samples=num_samples,
                                                block_size=block_size,
                                                index_length=len(prices.index),
                                                seed=SEED,
                                                init_to_end=False)
    bootstrap_prices.index = prices.index
    prices1 = pd.concat([bootstrap_prices, prices], axis=1)

    colors = num_samples * ['gray']
    colors1 = colors + ['red']
    kwargs = dict(x_date_freq='YE', legend_loc=None)
    with sns.axes_style("darkgrid"):
        fig, axs = plt.subplots(2, 1, figsize=(10, 7))
        qis.set_suptitle(fig, title='Log-performance and drawdowns of realized (red) and bootsrapped paths (grey)')
        figs.append(fig)
        qis.plot_prices_with_dd(prices=prices1,
                                is_log=True,
                                colors=colors1,
                                axs=axs,
                                **kwargs)

    log_returns = qis.to_returns(prices1, is_log_returns=True, is_first_zero=True)
    log_returns2 = np.square(log_returns)
    span = np.maximum(block_size, 5)
    ewma_vols = qis.compute_ewm_vol(data=log_returns,
                                    span=span,
                                    annualization_factor=252)
    with sns.axes_style("darkgrid"):
        fig, ax = plt.subplots(1, 1, figsize=(10, 7))
        qis.set_suptitle(fig, title=f"EWMA-{span} span volatility of realized (red) and bootsrapped paths (gray)")
        figs.append(fig)
        qis.plot_time_series(df=ewma_vols,
                             colors=colors1,
                             var_format='{:,.0%}',
                             ax=ax,
                             **kwargs)

    acfs, m_acf, std_acf = qis.estimate_acf_from_paths(log_returns, is_pacf=True, nlags=nlags)
    with sns.axes_style("darkgrid"):
        fig, ax = plt.subplots(1, 1, figsize=(10, 7))
        qis.set_suptitle(fig, title=f"Auto-correlation of returns of realized (red) and bootsrapped paths (grey)")
        figs.append(fig)
        qis.plot_line(df=acfs,
                      colors=colors1,
                      ax=ax,
                      **kwargs)

    acfs, m_acf, std_acf = qis.estimate_acf_from_paths(log_returns2, is_pacf=True, nlags=nlags)
    with sns.axes_style("darkgrid"):
        fig, ax = plt.subplots(1, 1, figsize=(10, 7))
        qis.set_suptitle(fig, title=f"Auto-correlation of squared returns of realized (red) and bootsrapped paths (grey)")
        figs.append(fig)
        qis.plot_line(df=acfs,
                      colors=colors1,
                      xlabel='lag',
                      ax=ax,
                      **kwargs)

    with sns.axes_style("darkgrid"):
        fig, ax = plt.subplots(1, 1, figsize=(10, 7))
        qis.set_suptitle(fig, title=f"Realized auto-correlation of squared returns (red) and Boxplot of auto-correlation of bootsrapped paths (grey)")
        figs.append(fig)
        qis.df_boxplot_by_index(df=acfs.drop(prices.name, axis=1),
                                legend_loc=None,
                                colors=colors,
                                xlabel='lag',
                                showmedians=True,
                                ax=ax)
        # add reasized labels
        qis.add_scatter_points(label_x_y=[(lag, v) for lag, v in acfs[prices.name].to_dict().items()],
                               color='red',
                               ax=ax)
        qis.set_legend(ax=ax,
                       labels=[prices.name],
                       markers=["*"],
                       colors=['red'],
                       handlelength=0)

        return figs


def plot_autocorr_in_block_size(prices: pd.Series,
                                block_sizes: List[int] = (5, 10, 20, 60),
                                num_samples: int = 1000,
                                nlags: int = 20,
                                ) -> List[plt.Figure]:
    """
    use different block_size to check for autocorrelation
    """
    figs = []
    for idx, block_size in enumerate(block_sizes):
        bootstrap_prices = qis.bootstrap_price_data(prices=prices,
                                                    bootsrap_type=qis.BootsrapType.STATIONARY,
                                                    bootsrap_output=qis.BootsrapOutput.SERIES_TO_DF,
                                                    num_samples=num_samples,
                                                    block_size=block_size,
                                                    index_length=len(prices.index),
                                                    seed=SEED,
                                                    init_to_end=False)
        bootstrap_prices.index = prices.index
        prices1 = pd.concat([bootstrap_prices, prices], axis=1)
        log_returns = qis.to_returns(prices1, is_log_returns=True, is_first_zero=True)
        log_returns2 = np.square(log_returns)
        acfs, m_acf, std_acf = qis.estimate_acf_from_paths(log_returns2, is_pacf=True, nlags=nlags)

        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(14, 10))
            qis.set_suptitle(fig, title=f"Realized auto-correlation of squared returns (red) and Boxplot of "
                                        f"auto-correlation of bootsrapped paths (grey) for block_size={block_size:0.0f}")
            figs.append(fig)
            qis.df_boxplot_by_index(df=acfs.drop(prices.name, axis=1),
                                    legend_loc=None,
                                    colors=num_samples * ['gray'],
                                    title=f"block_size={block_size:0.0f}",
                                    xlabel='lag',
                                    showmedians=True,
                                    ax=ax)
            # add reasized labels
            qis.add_scatter_points(label_x_y=[(lag, v) for lag, v in acfs[prices.name].to_dict().items()],
                                   color='red',
                                   ax=ax)
            qis.set_legend(ax=ax,
                           labels=[prices.name],
                           markers=["*"],
                           colors=['red'],
                           handlelength=0)
    return figs


class UnitTests(Enum):
    PLOT_BOOTSRAPPED_PRICES = 1
    PLOT_AUTOCORR_BLOCKSIZES = 2


def run_unit_test(unit_test: UnitTests):
    # to save pds
    LOCAL_PATH = "C://Users//artur//OneDrive//analytics//outputs//"

    # download spy prices
    prices = yf.download(tickers=['SPY'], start=None, end=None, ignore_tz=True)['Close'].rename('Realised')

    if unit_test == UnitTests.PLOT_BOOTSRAPPED_PRICES:
        # use small number of num_samples for illustration
        figs = plot_bootsrap_paths(prices=prices,
                                   block_size=30,
                                   num_samples=50)
        qis.save_figs_to_pdf(figs, file_name='bootstrap_illustrations', local_path=LOCAL_PATH)

    elif unit_test == UnitTests.PLOT_AUTOCORR_BLOCKSIZES:
        # block_size = 1 corresponds to iid sampling
        figs = plot_autocorr_in_block_size(prices=prices,
                                           block_sizes=[1, 2, 5, 10, 20, 40, 60, 120, 180],
                                           num_samples=1000)
        qis.save_figs_to_pdf(figs, file_name='bootstrap_block_sizes', local_path=LOCAL_PATH)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PLOT_AUTOCORR_BLOCKSIZES

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
generate and load test data using yf library
"""

import pandas as pd
import yfinance as yf
from enum import Enum

import qis.file_utils as fu
import qis.local_path as local_path


RESOURCE_PATH = local_path.get_paths()['RESOURCE_PATH']


def load_etf_data() -> pd.DataFrame:
    prices = fu.load_df_from_csv(file_name='etf_prices', local_path=RESOURCE_PATH)
    return prices


class UnitTests(Enum):
    ETF_PRICES = 1
    TEST_LOADING = 2


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.ETF_PRICES:
        prices = yf.download(tickers=['SPY', 'QQQ', 'EEM', 'TLT', 'IEF', 'LQD', 'HYG', 'SHY', 'GLD'],
                             start=None, end=None,
                             ignore_tz=True)['Close']
        print(prices)
        fu.save_df_to_csv(df=prices, file_name='etf_prices', local_path=RESOURCE_PATH)

    elif unit_test == UnitTests.TEST_LOADING:
        prices = load_etf_data()
        print(prices)


if __name__ == '__main__':

    unit_test = UnitTests.ETF_PRICES

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




from qis.models.linear.auto_corr import (
    compute_path_lagged_corr,
    compute_path_autocorr,
    compute_autocorr_df,
    compute_ewm_matrix_autocorr,
    compute_ewm_matrix_autocorr_df,
    estimate_acf_from_path,
    estimate_acf_from_paths,
    compute_ewm_vector_autocorr,
    compute_ewm_vector_autocorr_df
)

from qis.models.linear.corr_cov_matrix import (
    CorrMatrixOutput,
    estimate_rolling_ewma_covar,
    compute_ewm_corr_df,
    compute_ewm_corr_single,
    compute_masked_covar_corr,
    corr_to_pivot_row,
    matrix_regularization,
    compute_path_corr
)

from qis.models.linear.ewm import (
    InitType,
    MeanAdjType,
    CrossXyType,
    NanBackfill,
    ewm_recursion,
    compute_ewm,
    compute_ewm_long_short,
    compute_ewm_long_short_filter,
    compute_ewm_alpha_r2_given_prediction,
    compute_ewm_beta_alpha_forecast,
    compute_ewm_cross_xy,
    compute_ewm_covar,
    compute_ewm_covar_tensor,
    compute_ewm_covar_tensor_vol_norm_returns,
    compute_ewm_sharpe,
    compute_ewm_sharpe_from_prices,
    compute_ewm_std1_norm,
    compute_ewm_vol,
    compute_ewm_newey_west_vol,
    compute_ewm_xy_beta_tensor,
    compute_one_factor_ewm_betas,
    compute_roll_mean,
    compute_rolling_mean_adj,
    set_init_dim1,
    set_init_dim2
)

from qis.models.linear.ewm_convolution import ConvolutionType, SignalAggType, ewm_xy_convolution

from qis.models.linear.ewm_factors import (LinearModel,
                                           EwmLinearModel,
                                           compute_portfolio_benchmark_betas,
                                           compute_portfolio_benchmark_beta_alpha_attribution,
                                           compute_benchmarks_beta_attribution,
                                           estimate_linear_model)

from qis.models.linear.pca import(
    compute_eigen_portfolio_weights,
    apply_pca,
    compute_data_pca_r2,
    compute_pca_r2
)

from qis.models.linear.plot_correlations import(
    plot_returns_corr_matrix_time_series,
    plot_returns_corr_table,
    plot_returns_ewm_corr_table
)

from qis.models.linear.ra_returns import(
    ReturnsTransform,
    compute_ewm_ra_returns_momentum,
    compute_ra_returns,
    compute_ewm_long_short_filtered_ra_returns,
    map_signal_to_weight,
    SignalMapType,
    compute_returns_transform,
    compute_rolling_ra_returns,
    compute_sum_freq_ra_returns,
    compute_sum_rolling_ra_returns,
    get_paired_rareturns_signals
)

from qis.models.stats.bootstrap import (
    BootsrapOutput,
    BootsrapType,
    bootstrap_ar_process,
    bootstrap_data,
    bootstrap_price_data,
    bootstrap_price_fundamental_data,
    compute_ar_residuals,
    generate_bootstrapped_indices
)

from qis.models.stats.ohlc_vol import (
    OhlcEstimatorType,
    estimate_hf_ohlc_vol,
    estimate_ohlc_var
)

from qis.models.linear.ewm_winsor_outliers import (
    ReplacementType,
    OutlierPolicy,
    OutlierPolicyTypes,
    filter_outliers,
    ewm_insample_winsorising,
    compute_ewm_score
)

from qis.models.stats.rolling_stats import (RollingPerfStat,
                                            compute_rolling_perf_stat)



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import qis as qis
import yfinance as yf

tickers_weights = dict(SPY=0.6, IEF=0.4)
tickers = list(tickers_weights.keys())
prices = yf.download(tickers, start=None, end=None)['Close'][tickers]
prices = prices.asfreq('B', method='ffill').dropna()

balanced_60_40a = qis.backtest_model_portfolio(prices=prices, weights=tickers_weights, rebalancing_freq='QE',
                                              ticker='Zero Cost').get_portfolio_nav()
balanced_60_40b = qis.backtest_model_portfolio(prices=prices, weights=tickers_weights, rebalancing_freq='QE',
                                              ticker='2% Cost',
                                               management_fee=0.02).get_portfolio_nav()
navs = pd.concat([balanced_60_40a, balanced_60_40b], axis=1)
qis.plot_prices_with_dd(prices=navs)

plt.show()



"""
implement group by operations on df
"""
# packages
import numpy as np
import pandas as pd
from typing import Union, List, Dict, Callable, Optional
from enum import Enum

# qis
import qis.utils.struct_ops as sop
import qis.utils.df_agg as dfa


def get_group_dict(group_data: pd.Series,
                   index_data: Union[pd.Index, List[str], np.ndarray] = None,
                   group_order: List[str] = None,
                   total_column: Union[str, None] = None
                   ) -> Dict[str, List[str]]:
    """
    take group_data and return dictionary of index tickers mapped to group_data
    with index_data can do double mappings group_data to key, then key:index_data
    """
    if not isinstance(group_data, pd.Series):
        raise ValueError(f"group_data type={type(group_data)} must be series")

    if np.any(group_data.isnull()):
        raise ValueError(f"group data is nan\ngroup_data=\n{group_data}:"
                         f"\n{group_data.loc[np.isnan(group_data.isnull())]}")

    if index_data is not None:
        if isinstance(index_data, list) or isinstance(index_data, np.ndarray):  # convert to Index
            index_data = pd.Index(index_data)
        elif not isinstance(index_data, pd.Index):
            raise TypeError(f"unsupported type {type(index_data)}")
        if np.any(index_data.isnull()):
            raise ValueError(f"index_data is nan:\n{index_data[index_data.isnull()]}")

    group_data_ls = group_data.to_list()
    data_groups = list(np.unique(group_data_ls))
    if group_order is not None:
        groups = sop.merge_lists_unique(list1=group_order, list2=data_groups)
    else:
        groups = data_groups

    # fill group dict
    group_dict = {}
    if total_column is not None:
        group_tickers = group_data.index
        if index_data is not None:  # match group_tickers in index_data: necessary if index_data != group_data.index
            group_tickers = index_data[np.isin(index_data, group_tickers, assume_unique=True)]
        group_dict[total_column] = group_tickers

    # match group to instruments in group_data
    for group in groups:
        if group in data_groups:
            group_instruments = group_data.index[np.isin(group_data_ls, [group], assume_unique=True)]
            if index_data is not None:  # match group_tickers in index_data: necessary if index_data != group_data.index
                group_instruments = index_data[np.isin(index_data, group_instruments, assume_unique=True)]
            group_dict[group] = group_instruments.to_list()

    return group_dict


def split_df_by_groups(df: pd.DataFrame,
                       group_data: pd.Series,
                       total_column: Union[str, None] = None,
                       group_order: List[str] = None
                       ) -> Dict[str, pd.DataFrame]:
    """
    take pandas data which are index=time series, data = assets
    slit data according to grouping by rows in group_data[asset_group_column] in group_data
    the match of columns in data to rows in group_data is done using group_data.index or group_data[ticker_column_to_match]
    """
    # group by descriptive data
    group_dict = get_group_dict(group_data=group_data,
                                index_data=df.columns.to_numpy(),
                                group_order=group_order,
                                total_column=total_column)

    grouped_data = {}
    for key, columns in group_dict.items():
        if len(columns) > 0:  # get subset of pandas by columns, copy() to break view
            grouped_data[key] = df[columns].copy()

    return grouped_data


def agg_df_by_groups_ax1(df: pd.DataFrame,
                         group_data: pd.Series,
                         agg_func: Callable[[pd.DataFrame], pd.Series] = np.nansum,
                         total_column: Union[str, None] = None,
                         group_order: List[str] = None
                         ) -> pd.DataFrame:
    """
    take df[index=time series, data = asset data]
    slit data according to grouping by rows in group_data[asset_group_column] in group_data
    agg data by groups
    """
    # group by descriptive data
    group_dict = get_group_dict(group_data=group_data,
                                index_data=df.columns.to_numpy(),
                                group_order=group_order,
                                total_column=total_column)

    grouped_data = {}
    for key, columns in group_dict.items():
        if len(columns) > 0:  # get subset of pandas by columns, copy() to break view
            grouped_data[key] = df[columns].apply(agg_func, axis=1).rename(key)
    grouped_data = pd.DataFrame.from_dict(grouped_data, orient='columns')
    return grouped_data


def agg_df_by_groups(df: pd.DataFrame,
                     group_data: pd.Series,
                     group: str = None,
                     agg_func: Callable[[pd.DataFrame], pd.Series] = dfa.nansum,
                     total_column: Union[str, None] = None,
                     is_total_first: bool = True,
                     group_order: List[str] = None,
                     axis: Optional[int] = 1  # axis for integration
                     ) -> pd.DataFrame:
    """
    take pandas data which are index=time series, data = instruments
    slit data according to grouping by rows in group_data[group_column] in group_data
    the match of columns in data to rows in group_data is done using group_data.index or group_data[ticker_column_to_match]
    agg_func: pd.DataFrame -> Union[pd.Series, pd.DataFrame] is function for aggregation
    """
    # group by descriptive data
    group_dict = get_group_dict(group_data=group_data,
                                index_data=df.columns.to_numpy(),
                                group_order=group_order)

    # apply agg_func to grouped pandas
    agg_grouped_datas = []
    if group is None:  # output pandas with columns by dictionary keys
        for key, group_columns in group_dict.items():
            if len(group_columns) > 0:
                agg_grouped_datas.append(agg_func(df[group_columns], axis=axis).rename(key))
        agg_grouped_data = pd.concat(agg_grouped_datas, axis=1)

    else:  # just take the group
        if group in group_dict.keys():
            agg_grouped_data = agg_func(df[group_dict[group]]).to_frame(name=group)
        else:
            raise ValueError(f"{group} in keys {group_dict.keys()}")

    # insert total
    if total_column is not None:
        if not isinstance(total_column, str):
            raise TypeError(f"in agg_data_by_groups: {total_column} must be string")
        total = agg_func(df)
        if is_total_first:
            agg_grouped_data.insert(loc=0, column=total_column, value=total.values)
        else:
            agg_grouped_data.insert(loc=len(agg_grouped_data.columns), column=total_column, value=total.values)
    return agg_grouped_data


def agg_df_by_group_with_avg(df: pd.DataFrame,
                             group_data: pd.Series,
                             group_order: List[str] = None,
                             agg_func: Callable = dfa.nanmean,
                             agg_func_id: str = 'mean',
                             total_column: str = 'Universe mean'
                             ) -> Dict[str, pd.DataFrame]:
    """
    create ac grouped data dict with universes median and group medians
    """
    grouped_data = split_df_by_groups(df=df,
                                      group_data=group_data,
                                      group_order=group_order,
                                      total_column=None)

    group_avg = agg_df_by_groups(df=df,
                                 group_data=group_data,
                                 agg_func=agg_func,
                                 total_column=total_column)

    # add group median to grouped_data
    grouped_data_avg = dict()
    grouped_data_avg[total_column] = group_avg
    for ac_id, ac_data in grouped_data.items():
        data_avg = group_avg[ac_id].rename(f"{ac_id} {agg_func_id}")
        ac_data = pd.concat([data_avg, ac_data], axis=1)
        grouped_data_avg[ac_id] = ac_data

    return grouped_data_avg


def fill_df_with_group_avg(df: pd.DataFrame,
                           group_data: pd.Series,
                           agg_func: Callable[[pd.DataFrame], pd.Series] = dfa.nanmean,
                           group_order: List[str] = None
                           ) -> pd.DataFrame:
    """
    compute group avg and return data for each column with corresponding group avg
    """
    group_avg = agg_df_by_groups(df=df,
                                 group_data=group_data,
                                 group_order=group_order,
                                 agg_func=agg_func,
                                 total_column=None)
    avg_data = {}
    for column in df.columns:
        avg_data[column] = group_avg[group_data[column]]
    avg_data = pd.DataFrame.from_dict(avg_data, orient='columns')
    return avg_data


def sort_df_by_index_group(df: pd.DataFrame,
                           group_column: str,
                           sort_column: str,
                           ascending: bool = False,
                           group_order: List[str] = None
                           ) -> pd.DataFrame:

    group_dict = get_group_dict(group_data=df[group_column],
                                group_order=group_order)
    sorted_datas = []
    for group, tickers in group_dict.items():
        sorted_datas.append(df.loc[tickers, :].sort_values(by=[sort_column], ascending=ascending))
    sorted_data = pd.concat(sorted_datas, axis=0)

    return sorted_data


def set_group_loadings(group_data: pd.Series, group_order: List[str] = None) -> pd.DataFrame:
    if not isinstance(group_data, pd.Series):
        raise ValueError(f"{type(group_data)} must be pd.Series")
    if group_order is None:
        group_order = group_data.unique()
    group_loadings = {}
    for group in group_order:
        group_loadings[group] = pd.Series(np.where(group_data == group, 1.0, 0.0), index=group_data.index)
    group_loadings = pd.DataFrame.from_dict(group_loadings, orient='columns')
    return group_loadings


def convert_df_column_to_df_by_groups(df: pd.DataFrame,
                                      group_data: pd.Series,
                                      column: str,
                                      total_column: Union[str, None] = None,
                                      group_order: List[str] = None
                                      ) -> pd.DataFrame:
    """
    df is dataframe with index = assets
    group_data is series with index = assets and value = groups
    take dataframe column: split it to groups and produce df of values with columns = groups
    the index value is lost
    """
    # create df dict
    dfs = split_df_by_groups(df=df[column].to_frame().T, group_data=group_data,
                             total_column=total_column,
                             group_order=group_order)
    dfs1 = {}
    for key, df in dfs.items():
        dfs1[key] = df.T.reset_index(drop=True)[column]  # make series from df
    df1 = pd.DataFrame.from_dict(dfs1, orient='columns')
    return df1


class UnitTests(Enum):
    GROUP = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.GROUP:

        group_data = pd.Series(dict(SPY='Equities', QQQ='Equities', EEM='Equities', TLT='Bonds',
                                    IEF='Bonds', SHY='Bonds', LQD='Credit', HYG='HighYield', GLD='Gold'))

        group_dict = get_group_dict(group_data=group_data)
        print(f"group_dict=\n{group_dict}")

        group_dict_ordered = get_group_dict(group_data=group_data, group_order=list(group_data.unique()))
        print(f"group_dict_ordered=\n{group_dict_ordered}")

        group_dict_subset = get_group_dict(group_data=group_data,
                                           index_data=group_data.index[:5].to_list(),
                                           group_order=list(group_data.unique()))
        print(f"group_dict_subset=\n{group_dict_subset}")


if __name__ == '__main__':

    unit_test = UnitTests.GROUP

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# packages
import pandas as pd
import matplotlib.pyplot as plt
from enum import Enum
import yfinance as yf
import qis


class UnitTests(Enum):
    RETURNS_BOXPLOT = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.RETURNS_BOXPLOT:
        asset = 'SPY'
        regime_benchmark = '^VIX'
        prices = yf.download([asset, regime_benchmark], start=None, end=None)['Close'].dropna()
        prices = prices.asfreq('W-FRI', method='ffill')
        data = pd.concat([prices[asset].pct_change(),  # use returns over (t_0, t_1]
                          prices[regime_benchmark].shift(1)  # use level at t_0
                          ], axis=1).dropna()

        qis.df_boxplot_by_classification_var(df=data,
                                             x=regime_benchmark,
                                             y=asset,
                                             num_buckets=4,
                                             x_hue_name='VIX bucket',
                                             title=f"{asset} returns conditional on {regime_benchmark}",
                                             xvar_format='{:.2f}',
                                             yvar_format='{:.2%}',
                                             add_xy_mean_labels=True)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.RETURNS_BOXPLOT

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




"""
correlation related core
"""
# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from enum import Enum
from typing import Tuple, List, Union, Optional, Dict
from numba import njit

# qis
import qis.utils.dates as da
import qis.plots.time_series as pts
import qis.models.linear.ewm as ewm
import qis.perfstats.returns as ret


def estimate_rolling_ewma_covar(prices: pd.DataFrame,
                                time_period: da.TimePeriod,  # when we start estimation
                                returns_freq: str = 'W-WED',
                                rebalancing_freq: str = 'QE',
                                span: int = 52,
                                is_apply_vol_normalised_returns: bool = False,
                                demean: bool = True,
                                apply_an_factor: bool = True
                                ) -> Dict[pd.Timestamp, pd.DataFrame]:
    """
    compute ewma covar matrix: supporting for nans in prices
    covar data frequency is rebalancing_freq for period time_period
    output is dict[estimation timestamp, pd.Dataframe(estimated_covar)
    """
    returns = ret.to_returns(prices=prices, is_log_returns=True, drop_first=True, freq=returns_freq)
    returns_np = returns.to_numpy()
    if demean:
        x = returns_np - ewm.compute_ewm(returns_np, span=span)
    else:
        x = returns_np

    if is_apply_vol_normalised_returns:
        covar_tensor_txy, _, _ = ewm.compute_ewm_covar_tensor_vol_norm_returns(a=x, span=span, nan_backfill=ewm.NanBackfill.ZERO_FILL)
    else:
        covar_tensor_txy = ewm.compute_ewm_covar_tensor(a=x, span=span, nan_backfill=ewm.NanBackfill.ZERO_FILL)

    # create rebalancing schedule
    rebalancing_schedule = da.generate_rebalancing_indicators(df=returns, freq=rebalancing_freq)

    tickers = prices.columns.to_list()
    covars = {}
    if apply_an_factor:
        an_factor = da.infer_an_from_data(data=returns)
    else:
        an_factor = 1.0
    start_date = time_period.start.tz_localize(tz=returns.index.tz)  # make sure tz is alined with rebalancing_schedule
    for idx, (date, value) in enumerate(rebalancing_schedule.items()):
        if value and date >= start_date:
            covar_t = pd.DataFrame(covar_tensor_txy[idx], index=tickers, columns=tickers)
            covars[date] = an_factor*covar_t
    return covars


@njit
def compute_path_corr(a1: np.ndarray,
                      a2: np.ndarray
                      ) -> np.ndarray:
    """
    compute paths correlation between columns of a1 and a2
    """
    is_1d = (a1.ndim == 1)
    if not is_1d:
        ncols = a1.shape[1]
        acorr = np.zeros(ncols)
        for idx in range(ncols):
            acorr[idx] = np.corrcoef(a1[:, idx], a2[:, idx], rowvar=False)[0][1]
    else:
        acorr = np.corrcoef(a1, a2, rowvar=False)[0][1]

    return acorr


def compute_masked_covar_corr(data: Union[np.ndarray, pd.DataFrame],
                              is_covar: bool = True,
                              bias: bool = False
                              ) -> Union[np.ndarray, pd.DataFrame]:
    """
    given returns: time * assets
    compute covar by masking nans
    """
    if isinstance(data, pd.DataFrame):
        data_np = data.to_numpy()
    elif isinstance(data, np.ndarray):
        data_np = data
    else:
        raise ValueError(f"unsuported type {type(data)}")

    if np.any(np.isnan(data_np)):  # applay masked arrays
        if is_covar:
            covar = np.ma.cov(np.ma.masked_invalid(data_np), rowvar=False, bias=bias, allow_masked=True).data
        else:
            covar = np.ma.corrcoef(np.ma.masked_invalid(data_np), rowvar=False, bias=bias, allow_masked=True).data
    else:
        if is_covar:
            covar = np.cov(data_np, rowvar=False, bias=bias)
        else:
            covar = np.corrcoef(data_np, rowvar=False, bias=bias)

    if isinstance(data, pd.DataFrame):
        covar = pd.DataFrame(data=covar, index=data.columns, columns=data.columns)

    return covar


def corr_to_pivot_row(pivot: np.ndarray,
                      data: np.ndarray,
                      is_normalized: bool = True,
                      vol_scalers: List[Tuple[float, float]] = None  # [0] is pivot vol, [1] vol of asset
                      ) -> np.ndarray:
    """
    compute correlation row between pivot row of returns and return columns in data
    pivot returns are row data (r1,r2)
    data columns are column data ([c1, c2])
    output is correlation row of pivot to columns
    columns may have nans but pivot must be non-nan
    """
    n = len(data[0])
    corrs = np.zeros(n)

    # split column data into columns arrays
    column_data = np.hsplit(data, n)
    for idx, column_data in enumerate(column_data):

        # columnn data may have nans different from pivot
        # cross wil get nans from column_data
        # need to transpose column_data back to rows data
        cross = pivot * column_data.T

        if vol_scalers is not None:
            # multiply by n of non nans
            num = np.count_nonzero(np.isnan(cross) == False)
            std2 = num*vol_scalers[idx][0]*vol_scalers[idx][1]

        else:
            if is_normalized:
                cond = np.isnan(cross) == False  # cond will be [[]] array
                if np.any(cond == True):
                    clean_pivot = pivot[cond[0]]
                    clean_column = column_data[cond[0]]
                else:
                    clean_pivot = pivot
                    clean_column = column_data
                std2 = np.sqrt(np.nansum(clean_pivot*clean_pivot) * np.nansum(clean_column*clean_column))

            else:
                std2 = 1.0

        num_sum = np.nansum(cross)
        if not np.isnan(std2) and not np.isclose(std2, 0.0) and not np.isnan(num_sum):
            corrs[idx] = num_sum / std2
        else:
            corrs[idx] = np.nan

    return corrs


class CorrMatrixOutput(Enum):
    FULL = 1
    TOP_ROW = 2
    SUB_TOP = 3


def compute_ewm_corr_df(df: pd.DataFrame,
                        corr_matrix_output: CorrMatrixOutput = CorrMatrixOutput.FULL,
                        span: Union[int, np.ndarray] = None,
                        ewm_lambda: float = 0.94,
                        init_value: np.ndarray = None,
                        init_type: ewm.InitType = ewm.InitType.ZERO
                        ) -> pd.DataFrame:
    """
    compute ewm corr as and output as xi-xj pandas j>i, i = 0,..
    """
    if init_value is None:
        init_value = ewm.set_init_dim2(data=df.to_numpy(), init_type=init_type)

    corr = ewm.compute_ewm_covar_tensor(a=df.to_numpy(),
                                        span=span,
                                        ewm_lambda=ewm_lambda,
                                        is_corr=True,
                                        covar0=init_value)
    corr_ijs = []
    for idx_i, column_i in enumerate(df.columns):
        if corr_matrix_output == CorrMatrixOutput.SUB_TOP and idx_i == 0:  # skip for idx_i = 0
            continue

        for idx_j, column_j in enumerate(df.columns):
            if corr_matrix_output == CorrMatrixOutput.TOP_ROW:  # get j after i
                if idx_j > idx_i:
                    corr_ij = pd.Series(corr[:, idx_i, idx_j], name=f"{column_i} - {column_j}")
                    corr_ijs.append(corr_ij)
            else:  # get j before i
                if idx_j < idx_i:
                    corr_ij = pd.Series(corr[:, idx_i, idx_j], name=f"{column_i} - {column_j}")
                    corr_ijs.append(corr_ij)
        if corr_matrix_output == CorrMatrixOutput.TOP_ROW:  # stop after idx_i = 0
            break

    corrs_by_column = pd.concat(corr_ijs, axis=1)
    corrs_by_column = corrs_by_column.set_index(df.index)

    return corrs_by_column


def compute_ewm_corr_single(returns: pd.DataFrame,
                            ewm_lambda: float = 0.94,
                            span: Optional[int] = None,
                            time_period: da.TimePeriod = None
                            ) -> pd.Series:
    """
    plot correlation all time series in correlation matrix  as row
    """
    if len(returns.columns) != 2:
        raise ValueError("should be two columns {returns.columns}")

    if span is not None:
        ewm_lambda = 1.0 - 2.0 / (1.0 + span)

    corr = compute_ewm_corr_df(df=returns,
                               corr_matrix_output=CorrMatrixOutput.SUB_TOP,
                               ewm_lambda=ewm_lambda)

    if time_period is not None:
        corr = time_period.locate(corr)

    return corr.iloc[:, 0]


def matrix_regularization(covar: np.ndarray, cut: float = 1e-5) -> np.ndarray:
    eig_vals, eig_vecs = np.linalg.eigh(covar)
    eig_vals_alpha = np.where(np.greater(eig_vals, cut), eig_vals, 0.0)
    covar_a = eig_vecs @ np.diag(eig_vals_alpha) @ eig_vecs.T
    return covar_a


class UnitTests(Enum):
    CORR = 1
    EWMA_CORR_MATRIX = 2
    PLOT_CORR_MATRIX = 3
    MATRIX_REGULARIZATION = 4


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.CORR:
        t = 100
        n = 4
        data = np.random.normal(0, 1.0, (t, n))
        pivot = data[:, 0]
        corrs = corr_to_pivot_row(pivot=pivot, data=data, is_normalized=True)
        print(corrs)

    elif unit_test == UnitTests.EWMA_CORR_MATRIX:
        dates = pd.date_range(start='12/31/2018', end='12/31/2019', freq='B')
        n = 3
        mean = [-2.0, -1.0, 0.0]
        returns = pd.DataFrame(data=np.random.normal(mean, 1.0, (len(dates), n)),
                               index=dates,
                               columns=['x' + str(m + 1) for m in range(n)])

        corr = ewm.compute_ewm_covar_tensor(a=returns.to_numpy(), is_corr=True)
        print('corr')
        print(corr)
        print('corr_00')
        print(corr[0, 0, :])
        print('corr_01')
        print(corr[0, 1, :])
        print('corr_02')
        print(corr[0, 2, :])
        print('corr_last')
        print(corr[:, :, -1])

    elif unit_test == UnitTests.PLOT_CORR_MATRIX:

        dates = pd.date_range(start='31Dec2020', end='31Dec2021', freq='B')
        n = 3
        mean = [-2.0, -1.0, 0.0]
        returns = pd.DataFrame(data=np.random.normal(mean, 1.0, (len(dates), n)),
                               index=dates,
                               columns=['x' + str(m + 1) for m in range(n)])
        corrs = compute_ewm_corr_df(df=returns)
        print(corrs)

        pts.plot_time_series(df=corrs,
                             legend_stats=pts.LegendStats.AVG_LAST,
                             trend_line=pts.TrendLine.AVERAGE)

    elif unit_test == UnitTests.MATRIX_REGULARIZATION:
        covar = np.array([[1.0, -0.01, 0.01],
                         [-0.01, 0.5, 0.005],
                         [0.01, 0.005, 0.0001]])
        covar_a = matrix_regularization(covar=covar)
        print(covar_a)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.MATRIX_REGULARIZATION

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)


"""
implement winsorizing of time series data using ewm
1. compute mean_t and vol_t
2. select x% of outliers defined by normalized score (x_t-mean_t) / vol_t
3. replace or trim outliers as specified
"""
# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from enum import Enum
from typing import Union, NamedTuple, Optional, Tuple
from qis.models.linear.ewm import compute_ewm, compute_ewm_vol


class ReplacementType(Enum):
    EWMA_MEAN = 1
    NAN = 2
    QUANTILES = 3


class OutlierPolicy(NamedTuple):
    """
    specify filtering policy params
    """
    abs_ceil: Optional[float] = None  # remove all above
    abs_floor: Optional[float] = None  # remove all below
    std_abs_ceil: Optional[float] = None  # > 0
    std_abs_floor: Optional[float] = None  # < 0
    std_ewm_ceil: Optional[float] = None  # >0
    std_ewm_floor: Optional[float] = None  # <0
    ewm_lambda: Union[float, np.ndarray] = 0.94
    is_log_transform: bool = False
    nan_replacement_type: ReplacementType = ReplacementType.NAN


class OutlierPolicyTypes(OutlierPolicy, Enum):
    """
    defined policy type
    """
    HARD_CEIL_POLICY = OutlierPolicy(abs_floor=0.0001,
                                     std_abs_ceil=10.0)

    RANGE_CEIL_POLICY = OutlierPolicy(abs_floor=0.0001,
                                      std_abs_ceil=10.0)

    SOFT_RANGE_CEIL_POLICY = OutlierPolicy(abs_floor=1e-8,
                                           std_ewm_ceil=10.0,
                                           std_ewm_floor=None,
                                           std_abs_ceil=10.0)

    SOFT_POSITIVE_LOG_POLICY = OutlierPolicy(abs_floor=1e-8,
                                             std_ewm_ceil=10.0,
                                             std_ewm_floor=None,
                                             is_log_transform=True)
    NONE = None


def filter_outliers(data: Union[pd.DataFrame, pd.Series, np.ndarray],
                    outlier_policy: OutlierPolicy
                    ) -> Union[pd.DataFrame, pd.Series, np.ndarray]:

    np.seterr(invalid='ignore')  # off warnings

    if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
        orig_data = data.to_numpy()
    elif isinstance(data, np.ndarray):
        orig_data = data
    else:
        raise TypeError('filter_outliers: unsupported data type')

    clean_data = orig_data.copy()
    #  keep track of nans - nans will be put back to output data
    non_nan_cond = np.isfinite(orig_data)

    # imnitial replacement is using nans
    nan_replacement = np.full_like(orig_data, np.nan, dtype=np.float64)

    # remove absolute outliers
    if outlier_policy.abs_ceil is not None:
        clean_data = np.where(np.greater(clean_data, outlier_policy.abs_ceil, where=non_nan_cond),
                              nan_replacement, clean_data)
    if outlier_policy.abs_floor is not None:
        clean_data = np.where(np.less(clean_data, outlier_policy.abs_floor, where=non_nan_cond),
                              nan_replacement, clean_data)

    # now apply log transform
    if outlier_policy.is_log_transform:
        if outlier_policy.abs_floor is None:
            raise TypeError('is_log_transform must be applied with abs_floor > 0')
        log_cond = np.greater(clean_data, 0.0, where=non_nan_cond)
        clean_data = np.log(clean_data, where=log_cond)
    else:
        log_cond = None

    # remove relative outliers to in-sample std
    if outlier_policy.std_abs_ceil is not None or outlier_policy.std_abs_floor is not None:
        nan_mean = np.nanmean(clean_data, axis=0)
        nan_std = np.nanstd(clean_data, axis=0)

        if outlier_policy.std_abs_ceil is not None:
            ceil = np.add(nan_mean, nan_std*outlier_policy.std_abs_ceil, where=non_nan_cond)
            clean_data = np.where(np.greater(clean_data, ceil, where=non_nan_cond), nan_replacement, clean_data)

        if outlier_policy.std_abs_floor is not None:
            floor = np.add(nan_mean, nan_std*outlier_policy.std_abs_floor, where=non_nan_cond)
            clean_data = np.where(np.less(clean_data, floor, where=non_nan_cond), nan_replacement, clean_data)

    # now rolling ewm outliers
    if outlier_policy.std_ewm_ceil is not None or outlier_policy.std_ewm_floor is not None:

        ewm_mean, score = compute_ewm_score(data=clean_data, ewm_lambda=outlier_policy.ewm_lambda)
        if outlier_policy.std_ewm_ceil is not None:
            clean_data = np.where(np.greater(score, outlier_policy.std_ewm_ceil, where=non_nan_cond),
                                  nan_replacement, clean_data)

        if outlier_policy.std_ewm_floor is not None:
            clean_data = np.where(np.less(score, outlier_policy.std_ewm_floor, where=non_nan_cond),
                                  nan_replacement, clean_data)
    if outlier_policy.is_log_transform:
        clean_data = np.exp(clean_data, where=log_cond)

    # implemented replacement type is EWMA mean
    if outlier_policy.nan_replacement_type == ReplacementType.EWMA_MEAN:
        ewm_mean, _ = compute_ewm_score(data=clean_data, ewm_lambda=outlier_policy.ewm_lambda)
        filtered_data = np.where(np.isfinite(clean_data), clean_data, ewm_mean)
    else:
        filtered_data = np.where(non_nan_cond, clean_data, nan_replacement)

    if isinstance(data, pd.DataFrame):
        filtered_data = pd.DataFrame(data=filtered_data, columns=data.columns, index=data.index)
    elif isinstance(data, pd.Series):
        filtered_data = pd.Series(data=filtered_data, name=data.name, index=data.index)

    return filtered_data


def ewm_insample_winsorising(data: Union[pd.DataFrame, pd.Series, np.ndarray],
                             ewm_lambda: Union[float, np.ndarray] = 0.94,
                             quantile_cut: float = 0.025,
                             nan_replacement_type: ReplacementType = ReplacementType.EWMA_MEAN
                             ) -> Union[pd.DataFrame, pd.Series, np.ndarray]:

    if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
        np_data = data.to_numpy()
    elif isinstance(data, np.ndarray):
        np_data = data.copy()
    else:
        raise TypeError('ewm_winsorising: unsupported data type')

    # 1 compute ewm score
    ewm_mean, score = compute_ewm_score(data=np_data, ewm_lambda=ewm_lambda)

    lower_quantile = np.quantile(score, quantile_cut, axis=0)
    upper_quantile = np.quantile(score, 1.0-quantile_cut, axis=0)
    # print(f"lower_quantile={lower_quantile}, upper_quantile={upper_quantile}")

    if nan_replacement_type == ReplacementType.EWMA_MEAN:
        replacement_cond = np.logical_or(score < lower_quantile, score > upper_quantile)
        winsor_data = np.where(replacement_cond, ewm_mean, np_data)

    elif nan_replacement_type == ReplacementType.NAN:
        replacement_cond = np.logical_or(score < lower_quantile, score > upper_quantile)
        winsor_data = np.where(replacement_cond, np.full_like(np_data, np.nan), np_data)

    elif nan_replacement_type == ReplacementType.QUANTILES:
        winsor_data = np.where(score < lower_quantile, np.quantile(np_data, quantile_cut, axis=0), np_data)
        winsor_data = np.where(score > upper_quantile, np.quantile(np_data, 1.0-quantile_cut, axis=0), winsor_data)
    else:
        raise TypeError('replacement_type not implemented')

    if isinstance(data, pd.DataFrame):
        winsor_data = pd.DataFrame(data=winsor_data, columns=data.columns, index=data.index)
    elif isinstance(data, pd.Series):
        winsor_data = pd.Series(data=winsor_data, name=data.name, index=data.index)

    return winsor_data


def compute_ewm_score(data: np.ndarray,
                      ewm_lambda: Union[float, np.ndarray] = 0.94,
                      is_clip: bool = True,
                      clip_quantile: float = 0.16
                      ) -> (np.ndarray, np.ndarray):

    ewm_mean = compute_ewm(data=data, ewm_lambda=ewm_lambda)
    ewm_vol = compute_ewm_vol(data=data, ewm_lambda=ewm_lambda)
    if is_clip:  # remove small values below 1 _ std quantile
        ewm_vol = np.clip(a=ewm_vol, a_min=np.nanquantile(ewm_vol, clip_quantile), a_max=None)
    non_nan_cond = np.isfinite(data)
    score = np.divide(np.subtract(data, ewm_mean), ewm_vol, where=non_nan_cond)
    return ewm_mean, score


# @njit
def ewm_winsdor_markovian_score(a: np.ndarray,
                                init_value: Union[float, np.ndarray],
                                init_var: Union[float, np.ndarray] = None,
                                score_threshold: float = 5.0,
                                span: Union[int, np.ndarray] = 31,
                                ewm_lambda: Union[float, np.ndarray] = None,
                                is_start_from_first_nonan: bool = True
                                ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:

    """
    use ewma score to filter out outliers in non-anticipating markovian way
    data: numpy with dimension = t*n

    score_t is defined as non-anticipating:
    score_t = (x[t]-ewm[t-1] / np.sqrt(ewm2[t-1])
    outlier x[t] is defined when:
    np.abs(score_t) > threshold

    if x[t] is outlier, it is ignored for ewm:
    ewm[t] = ewm[t-1]
    ewm2[t] = ewm2[t-1]
    else ewm is computed using recursion:
    ewm[t] = (1-lambda) * x[t] + lambda*ewm[t-1]
    ewm2[t] = (1-lambda) * (x[t]-ewm[t])^2 + lambda*ewm2[t-1]

    if x[t] is nan:
    ewm[t] = ewm[t-1]
    ewm2[t] = ewm2[t-1]

    assumption is that no np.nan value is returned from the function

    ewm_lambda: float or ndarray of dimension n
    init_value: initial value of dimension n
    start_from_first_nonan: start filling nans only from the first non-nan in underlying data: recomended because
                            it avoids backfilling of init_value
    """
    if span is not None:
        ewm_lambda = 1.0 - 2.0 / (span + 1.0)
    ewm_lambda_1 = 1.0 - ewm_lambda

    is_1d = (a.ndim == 1)  # or a.shape[1] == 1)

    # initialize all
    ewm = np.full_like(a, fill_value=np.nan, dtype=np.double)
    ewm2 = np.full_like(a, fill_value=np.nan, dtype=np.double)
    score = np.full_like(a, fill_value=np.nan, dtype=np.double)
    clean_a = np.full_like(a, fill_value=np.nan, dtype=np.double)

    if init_var is None:
        if is_1d:
            init_var = 0.1
        else:
            init_var = 0.1*np.ones(a.shape[1])

    if is_start_from_first_nonan:
        if is_1d:  # cannot use np.where
            last_ewm = init_value if np.isfinite(a[0]) else np.nan
        else:
            last_ewm = np.where(np.isfinite(a[0]), init_value, np.nan)
    else:
        last_ewm = init_value

    last_ewm2 = np.maximum(last_ewm * last_ewm, init_var)

    ewm[0] = last_ewm
    ewm2[0] = last_ewm2
    score[0] = 0.0

    # recurse from 1
    for t in np.arange(1, a.shape[0]):
        a_t = a[t]

        if is_start_from_first_nonan:
            # detect starting nonnans for when last ewma was np.nan and a_t is finite
            if is_1d:  # cannot use np.where
                if np.isfinite(last_ewm) == False and np.isfinite(a_t) == True:  # trick: if last_ewm is nan
                    last_ewm = init_value
                    last_ewm2 = np.maximum(init_value*init_value, init_var)
            else:
                new_nonnans = np.logical_and(np.isfinite(last_ewm) == False, np.isfinite(a_t) == True)
                if np.any(new_nonnans):
                    last_ewm = np.where(new_nonnans, init_value, last_ewm)
                    last_ewm2 = np.where(new_nonnans, np.maximum(init_value*init_value, init_var), last_ewm2)

        # fill nan-values
        current_ewm_ = ewm_lambda * last_ewm + ewm_lambda_1 * a_t
        current_ewm2_ = ewm_lambda * last_ewm2 + ewm_lambda_1 * np.square(a_t-current_ewm_)

        # score_t = np.divide(a_t - last_ewm, np.sqrt(last_ewm2), where=np.greater(last_ewm2, 0.0))
        score_t = np.where(np.greater(last_ewm2, 0.0), (a_t - last_ewm)/np.sqrt(last_ewm2), np.nan)
        is_outlier = np.abs(score_t) >= score_threshold

        if is_1d:   # np.where cannot be used
            if is_outlier:
                current_ewm = last_ewm
                current_ewm2 = last_ewm2
                clean_a_ = clean_a[t-1]
            else:
                current_ewm = current_ewm_
                current_ewm2 = current_ewm2_
                clean_a_ = a_t
        else:
            current_ewm = np.where(is_outlier, current_ewm_, last_ewm)
            current_ewm2 = np.where(is_outlier, current_ewm2_, last_ewm2)
            clean_a_ = np.where(is_outlier, clean_a[t-1], a_t)

        ewm[t] = last_ewm = current_ewm
        ewm2[t] = last_ewm2 = current_ewm2
        score[t] = score_t
        clean_a[t] = clean_a_

    return clean_a, ewm, ewm2, score


class UnitTests(Enum):
    TEST1 = 1
    MARKOVIAN_WINDSOR = 2


def run_unit_test(unit_test: UnitTests):

    np.random.seed(2) #freeze seed

    import qis as qis

    dates = pd.date_range(start='12/31/2018', end='12/31/2019', freq='B')
    n = 2

    data = pd.DataFrame(data=np.random.standard_t(df=2, size=(len(dates), n)),
                        index=dates,
                        columns=['x'+str(m+1) for m in range(n)])
    # data = pd.Series(data=data_np, index=dates, name='data')

    if unit_test == UnitTests.TEST1:
        winsor_data = ewm_insample_winsorising(data=data,
                                               ewm_lambda=0.94,
                                               nan_replacement_type=ReplacementType.EWMA_MEAN,
                                               quantile_cut=0.05)

        winsor_data.columns = [x + ' winsor' for x in data.columns]

        plot_data = pd.concat([data, winsor_data], axis=1)
        title = 'Data Winsor'

        qis.plot_time_series(df=plot_data,
                             title=title,
                             legend_loc='upper left',
                             legend_stats=qis.LegendStats.AVG,
                             last_label=qis.LastLabel.AVERAGE_VALUE,
                             trend_line=qis.TrendLine.AVERAGE,
                             var_format='{:.2f}')

    elif unit_test == UnitTests.MARKOVIAN_WINDSOR:
        clean_x, ewm, ewm2, score = ewm_winsdor_markovian_score(a=data.to_numpy(), span=7,
                                                                init_value=0.0,
                                                                init_var=0.1*np.ones(len(data.columns)))
        clean_x = pd.DataFrame(clean_x, index=dates, columns=[x + ' winsor' for x in data.columns])
        plot_data = pd.concat([data, clean_x], axis=1)
        qis.plot_time_series(df=plot_data,
                             title='Windsor',
                             legend_loc='upper left',
                             legend_stats=qis.LegendStats.AVG_STD_LAST,
                             var_format='{:.2f}')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.MARKOVIAN_WINDSOR

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
configuration for performance reports
"""
from enum import Enum
from typing import Dict, Any, Tuple, NamedTuple, Optional, List

import qis
from qis import PerfParams, BenchmarkReturnsQuantileRegimeSpecs, TimePeriod, PerfStat
import yfinance as yf

# default params have no risk-free rate
PERF_PARAMS = PerfParams(freq='W-WED', freq_reg='W-WED', alpha_an_factor=52, rates_data=None)
REGIME_PARAMS = BenchmarkReturnsQuantileRegimeSpecs(freq='QE')

PERF_COLUMNS_RF0 = (PerfStat.TOTAL_RETURN,
                    PerfStat.PA_RETURN,
                    PerfStat.VOL,
                    PerfStat.SHARPE_RF0,
                    PerfStat.MAX_DD,
                    PerfStat.MAX_DD_VOL,
                    PerfStat.SKEWNESS,
                    PerfStat.ALPHA_AN,
                    PerfStat.BETA,
                    PerfStat.R2,
                    PerfStat.ALPHA_PVALUE)


PERF_COLUMNS = (PerfStat.TOTAL_RETURN,
                PerfStat.PA_RETURN,
                PerfStat.VOL,
                PerfStat.SHARPE_RF0,
                PerfStat.SHARPE_EXCESS,
                PerfStat.MAX_DD,
                PerfStat.MAX_DD_VOL,
                PerfStat.SKEWNESS,
                PerfStat.ALPHA_AN,
                PerfStat.BETA,
                PerfStat.R2,
                PerfStat.ALPHA_PVALUE)


class ReportingFrequency(Enum):
    DAILY = 1
    MONTHLY = 2
    QUARTERLY = 3


class FactsheetConfig(NamedTuple):
    """
    enumerate key parameters for report comps and visuals on different time period or data frequency
    default is for daily data with at least 10y
    """
    freq: str = 'W-WED'  # for trading stats
    vol_freq: str = 'W-WED'
    vol_rolling_window: int = 13  # 12 span volatility of weekly returns
    freq_drawdown: str = 'B'
    freq_reg: str = 'W-WED'  # for beta regressions
    freq_var: str = 'B'  # for var computations
    var_span: float = 33.0 # for var computations
    alpha_an_factor: float = 52 # for W-WED returns
    freq_regime: str = 'QE'  # for regime frequency
    sharpe_rolling_window: int = 156  # 3y of weekly returns
    freq_sharpe: str = 'W-WED'  # for rolling sharpe
    turnover_rolling_period: int = 260  # turnover = turnover.rolling(turnover_roll_period).sum()
    freq_turnover: str = 'B'   # daily freq
    cost_rolling_period: int = 260  # turnover = turnover.rolling(turnover_roll_period).sum()
    freq_cost: Optional[str] = 'B'  # for rolling costs
    is_unit_based_traded_volume: bool = True  # for long-only protfolio use nrmalised costs
    factor_beta_span: int = 52  # to compute rolling beta
    freq_beta: str = 'W-WED'  # for scatter plot
    exposures_freq: str = 'W-WED'  # for plotting strategy exposures
    # general data
    perf_columns: List[PerfStat] = PERF_COLUMNS
    perf_stats_labels: List[PerfStat] = (PerfStat.PA_RETURN, PerfStat.VOL, PerfStat.SHARPE_RF0,)
    short: bool = True  # ra columns
    # next depend on report time period
    heatmap_freq: str = 'YE'  # for heatmap
    x_date_freq: str = 'YE'  # for time series plots
    date_format: str = '%b-%y'  # for dates


# create enumerations
FACTSHEET_CONFIG_DAILY_DATA_LONG_PERIOD = FactsheetConfig()

FACTSHEET_CONFIG_DAILY_DATA_SHORT_PERIOD = FactsheetConfig(heatmap_freq='ME',
                                                           x_date_freq='ME',
                                                           freq_regime='W-WED',
                                                           freq_reg='W-WED',
                                                           alpha_an_factor=52
                                                           )


FACTSHEET_CONFIG_MONTHLY_DATA_LONG_PERIOD = FactsheetConfig(freq='ME',
                                                            freq_drawdown='ME',
                                                            freq_reg='ME',
                                                            vol_freq='ME',
                                                            alpha_an_factor=12,
                                                            freq_regime='QE',
                                                            sharpe_rolling_window=36,
                                                            vol_rolling_window=13,
                                                            freq_sharpe='ME',
                                                            turnover_rolling_period=12,
                                                            freq_turnover='ME',
                                                            cost_rolling_period=12,
                                                            freq_cost='ME',
                                                            freq_var='ME',
                                                            var_span=12,
                                                            is_unit_based_traded_volume=True,
                                                            factor_beta_span=36,
                                                            freq_beta='ME',
                                                            exposures_freq='ME')

FACTSHEET_CONFIG_MONTHLY_DATA_SHORT_PERIOD = FACTSHEET_CONFIG_MONTHLY_DATA_LONG_PERIOD


FACTSHEET_CONFIG_QUARTERLY_DATA_LONG_PERIOD = FactsheetConfig(freq='QE',
                                                              freq_drawdown='ME',
                                                              freq_reg='QE',
                                                              vol_freq='QE',
                                                              alpha_an_factor=4,
                                                              freq_regime='QE',
                                                              sharpe_rolling_window=12,
                                                              vol_rolling_window=12,
                                                              freq_sharpe='QE',
                                                              turnover_rolling_period=4,
                                                              freq_turnover='QE',
                                                              cost_rolling_period=4,
                                                              freq_cost='QE',
                                                              freq_var='QE',
                                                              var_span=12,
                                                              is_unit_based_traded_volume=True,
                                                              factor_beta_span=12,
                                                              freq_beta='QE',
                                                              exposures_freq='QE')


def fetch_factsheet_config_kwargs(factsheet_config: FactsheetConfig = FACTSHEET_CONFIG_DAILY_DATA_LONG_PERIOD,
                                  add_rates_data: bool = True
                                  ) -> Dict[str, Any]:
    """
    return kwargs for factsheet reports
    """
    rates_data = None
    if add_rates_data:
        rates_data = yf.download('^IRX', start=None, end=None)['Close'].dropna() / 100.0
        if rates_data.empty:  # if online
            rates_data = None

    perf_params = PerfParams(freq=factsheet_config.freq,
                             freq_vol=factsheet_config.vol_freq,
                             freq_drawdown=factsheet_config.freq_drawdown,
                             freq_reg=factsheet_config.freq_reg,
                             alpha_an_factor=factsheet_config.alpha_an_factor,
                             rates_data=rates_data)
    regime_params = BenchmarkReturnsQuantileRegimeSpecs(freq=factsheet_config.freq_regime)
    kwargs = factsheet_config._asdict()

    if add_rates_data:
        kwargs['perf_stats_labels'] = (PerfStat.PA_RETURN, PerfStat.VOL, PerfStat.SHARPE_RF0, PerfStat.SHARPE_EXCESS,)
    else:
        kwargs['perf_columns'] = PERF_COLUMNS_RF0
        kwargs['perf_stats_labels'] = (PerfStat.PA_RETURN, PerfStat.VOL, PerfStat.SHARPE_RF0, )

    kwargs.pop('freq')  # remove frequency as some methods have default freq
    kwargs.update(dict(perf_params=perf_params, regime_params=regime_params))
    return kwargs


def fetch_default_perf_params() -> Tuple[PerfParams, BenchmarkReturnsQuantileRegimeSpecs]:
    """
    by default we use 3m US rate
    """
    rates_data = yf.download('^IRX', start=None, end=None)['Close'].dropna() / 100.0
    if rates_data.empty:  # if online
        rates_data = None
    perf_params = PerfParams(freq='W-WED', freq_reg='W-WED', rates_data=rates_data)
    regime_params = BenchmarkReturnsQuantileRegimeSpecs(freq='QE')

    return perf_params, regime_params


def fetch_default_report_kwargs(time_period: Optional[TimePeriod] = None,
                                reporting_frequency: ReportingFrequency = ReportingFrequency.DAILY,
                                long_threshold_years: float = 5.0,
                                add_rates_data: bool = True,
                                is_unit_based_traded_volume: bool = True
                                ) -> Dict[str, Any]:

    # use for number years > 5
    if time_period is not None:
        if time_period.get_time_period_an() > long_threshold_years:
            if reporting_frequency == ReportingFrequency.DAILY:
                factsheet_config = FACTSHEET_CONFIG_DAILY_DATA_LONG_PERIOD
            elif reporting_frequency == ReportingFrequency.QUARTERLY:
                factsheet_config = FACTSHEET_CONFIG_QUARTERLY_DATA_LONG_PERIOD
            else:  # default is monthly
                factsheet_config = FACTSHEET_CONFIG_MONTHLY_DATA_LONG_PERIOD
        else:
            if reporting_frequency == ReportingFrequency.DAILY:
                factsheet_config = FACTSHEET_CONFIG_DAILY_DATA_SHORT_PERIOD
            elif reporting_frequency == ReportingFrequency.QUARTERLY:
                factsheet_config = FACTSHEET_CONFIG_QUARTERLY_DATA_LONG_PERIOD
            else:  # default is monthly
                factsheet_config = FACTSHEET_CONFIG_MONTHLY_DATA_SHORT_PERIOD
    else:
        if reporting_frequency == ReportingFrequency.DAILY:
            factsheet_config = FACTSHEET_CONFIG_DAILY_DATA_LONG_PERIOD
        elif reporting_frequency == ReportingFrequency.QUARTERLY:
            factsheet_config = FACTSHEET_CONFIG_QUARTERLY_DATA_LONG_PERIOD
        else:
            factsheet_config = FACTSHEET_CONFIG_MONTHLY_DATA_LONG_PERIOD

    kwargs = fetch_factsheet_config_kwargs(factsheet_config=factsheet_config,
                                           add_rates_data=add_rates_data)
    kwargs = qis.update_kwargs(kwargs, dict(is_unit_based_traded_volume=is_unit_based_traded_volume))
    return kwargs


# for pybloqs
margin_top = 1.0
margin_bottom = 1.0
line_height = 0.99
font_family = 'Calibri'
KWARGS_SUPTITLE = {'title_wrap': True, 'text_align': 'center', 'color': 'blue', 'font_size': "12px", 'font-weight': 'normal',
                   'title_level': 1, 'line_height': 0.7, 'inherit_cfg': False,
                   'margin_top': 0, 'margin_bottom': 0,
                   'font-family': 'sans-serif'}
KWARGS_TITLE = {'title_wrap': True, 'text_align': 'left', 'color': 'blue', 'font_size': "12px",
                'title_level': 1, 'line_height': line_height, 'inherit_cfg': False,
                'margin_top': margin_top,  'margin_bottom': margin_bottom,
                'font-family': font_family}
KWARGS_DESC = {'title_wrap': True, 'text_align': 'left', 'font_size': "12px", 'font-weight': 'normal',
               'title_level': 2, 'line_height': line_height, 'inherit_cfg': False,
               'margin_top': margin_top, 'margin_bottom': margin_bottom,
               'font-family': font_family}
KWARGS_TEXT = {'title_wrap': True, 'text_align': 'left', 'font_size': "12px", 'font-weight': 'normal',
               'title_level': 2, 'line_height': line_height, 'inherit_cfg': False,
               'margin_top': margin_top, 'margin_bottom': margin_bottom,
               'font-family': font_family}
KWARGS_FIG = {'title_wrap': True, 'text_align': 'left', 'font_size': "12px",
              'title_level': 2, 'line_height': line_height, 'inherit_cfg': False,
              'margin_top': margin_top, 'margin_bottom': margin_bottom,
              'font-family': font_family}
KWARGS_FOOTNOTE = {'title_wrap': True, 'text_align': 'left', 'font_size': "12px", 'font-weight': 'normal',
                   'title_level': 3, 'line_height': line_height, 'inherit_cfg': False,
                   'margin_top': margin_top, 'margin_bottom': margin_bottom,
                   'font-family': font_family}



"""
compute seasonal strategy using monthly returns
"""

import numpy as np
import pandas as pd
import qis as qis


def q60(x):
    return x.quantile(0.6)


def q40(x):
    return x.quantile(0.4)


def compute_seasonal_signal(prices: pd.DataFrame) -> pd.DataFrame:
    """
    compute seasonal signal
    """
    prices = prices.asfreq('B', method='ffill')  # make B frequency
    returns = qis.to_returns(prices, freq='ME', drop_first=True)
    returns['month'] = returns.index.month
    seasonal_returns = returns.groupby('month').agg(['mean', q40, q60])
    signals = {}
    for asset in prices.columns:
        df = seasonal_returns[asset]
        signal = np.where(df['q40'] > 0.0, +1.0, np.where(df['q60'] < 0.0, -1.0, 0.0))
        signals[asset] = pd.Series(signal, index=df.index).fillna(0.0)
    signals = pd.DataFrame.from_dict(signals, orient='columns')
    return signals


def compute_rolling_seasonal_signals(prices: pd.DataFrame,
                                     num_sample_years: int = 25
                                     ) -> pd.DataFrame:
    rebalancing_years = list(np.arange(2000, 2025))
    monthly_returns = qis.to_returns(prices, freq='ME', drop_first=True, include_end_date=True)
    signals = []
    for year in rebalancing_years:
        # print(year)
        estimation_sample = prices.loc[f"{year-num_sample_years}":f"{year-1}", :]
        if year == 2024:
            print('here')
        # print(estimation_sample)
        investment_sample = monthly_returns.iloc[monthly_returns.index.year==year, :]
        investment_sample_index = investment_sample.index
        investment_sample_index = investment_sample_index.shift(-1, freq="ME")  # sift signal to end of previous month
        signal = compute_seasonal_signal(prices=estimation_sample)
        signal = signal.iloc[:len(investment_sample_index), :]  # drop last few mon
        # set signal of the investment sample with weight at the end of prior months
        signal = signal.set_index(investment_sample_index)
        # now move to start of the month
        signals.append(signal)
    signals = pd.concat(signals, axis=0)
    return signals



# packages
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import seaborn as sns
import qis as qis
from bbg_fetch import fetch_field_timeseries_per_tickers

prices = fetch_field_timeseries_per_tickers(tickers={'SPTR Index': 'SPTR'}, freq='B', field='PX_LAST').ffill()

yields = fetch_field_timeseries_per_tickers(tickers={'USGGBE10 Index': '10Y BE'}, freq='B', field='PX_LAST').ffill() / 100.0

roll_periodss = {'5y': 5*252, '10y': 10*252}
roll_periodss = {'10y': 10*252}
perfs = {}
for key, roll_periods in roll_periodss.items():
    perf, _ = qis.compute_rolling_perf_stat(prices=prices,
                                            rolling_perf_stat=qis.RollingPerfStat.PA_RETURNS,
                                            roll_freq='B',
                                            roll_periods=roll_periods)
    perfs[key] = perf.iloc[:, 0]
perfs = pd.DataFrame.from_dict(perfs, orient='columns').dropna(axis=0, how='all')

with sns.axes_style("darkgrid"):
    fig, ax = plt.subplots(1, 1, figsize=(8, 6), tight_layout=True)
    yvar_major_ticks1 = np.linspace(-0.05, 0.3, 8)
    yvar_major_ticks2 = np.linspace(0.0, 0.03, 6)
    qis.plot_time_series_2ax(df1=perfs,
                             df2=yields,
                             legend_stats=qis.LegendStats.AVG_LAST,
                             legend_stats2=qis.LegendStats.AVG_LAST,
                             title='Rolling 10y p.a. returns of SPTR Index vs US 10Y BE',
                             var_format='{:.1%}',
                             var_format_yax2='{:.1%}',
                             yvar_major_ticks1=yvar_major_ticks1,
                             yvar_major_ticks2=yvar_major_ticks2,
                             trend_line1=qis.TrendLine.ZERO_SHADOWS,
                             framealpha=0.9,
                             ax=ax)

plt.show()


# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import List, Tuple, Callable, Optional, Dict, Union
from enum import Enum
# qis
import qis.utils.dates as da
import qis.utils.df_str as dfs
import qis.utils.struct_ops as sop
import qis.perfstats.returns as ret
from qis.perfstats.config import PerfStat, PerfParams
import qis.perfstats.perf_stats as rpt
import qis.plots.scatter as psc
import qis.plots.table as ptb
import qis.plots.utils as put
from qis.plots.bars import plot_bars, plot_vbars


def get_ra_perf_columns(prices: Union[pd.DataFrame, pd.Series],
                        perf_params: PerfParams = None,
                        perf_columns: List[PerfStat] = rpt.STANDARD_TABLE_COLUMNS,
                        column_header: str = 'Asset',
                        df_to_add: pd.DataFrame = None,
                        is_to_str: bool = True,
                        **kwargs
                        ) -> pd.DataFrame:
    """
    compute ra perf table and get ra performance columns with data as string for tables
    """
    if isinstance(prices, pd.Series):
        prices = prices.to_frame()
    ra_perf_table = rpt.compute_ra_perf_table(prices=prices, perf_params=perf_params)
    data = pd.DataFrame(index=ra_perf_table.index)
    for perf_column in perf_columns:
        if perf_column.to_str() in ra_perf_table.columns:
            if is_to_str:
                data[perf_column.to_str()] = dfs.series_to_str(ds=ra_perf_table[perf_column.to_str()],
                                                               var_format=perf_column.to_format(**kwargs))
            else:
                data[perf_column.to_str()] = ra_perf_table[perf_column.to_str()]

    if df_to_add is not None:
        for idx, column in enumerate(df_to_add.columns):
            data.insert(idx, column=column, value=df_to_add[column].to_numpy())
    data.index.name = column_header
    return data


def plot_ra_perf_table(prices: Union[pd.DataFrame, pd.Series],
                       perf_params: PerfParams = None,
                       perf_columns: List[PerfStat] = rpt.STANDARD_TABLE_COLUMNS,
                       special_columns_colors: List[Tuple[int, str]] = ((0, 'blue'),),
                       rows_edge_lines: List[int] = None,
                       special_rows_colors: List[Tuple[int, str]] = None,
                       column_header: str = 'Asset',
                       fontsize: int = 10,
                       transpose: bool = False,
                       df_to_add: pd.DataFrame = None,
                       ax: plt.Subplot = None,
                       **kwargs
                       ) -> Optional[plt.Figure]:
    """
    plot ra perf table columns
    """
    df = get_ra_perf_columns(prices=prices,
                             perf_params=perf_params,
                             perf_columns=perf_columns,
                             column_header=column_header,
                             df_to_add=df_to_add,
                             **kwargs)
    fig = ptb.plot_df_table(df=df,
                            add_index_as_column=True,
                            transpose=transpose,
                            special_columns_colors=special_columns_colors,
                            special_rows_colors=special_rows_colors,
                            rows_edge_lines=rows_edge_lines,
                            fontsize=fontsize,
                            ax=ax,
                            **kwargs)
    return fig


def get_ra_perf_benchmark_columns(prices: pd.DataFrame,
                                  benchmark: str,
                                  drop_benchmark: bool = False,
                                  perf_params: PerfParams = None,
                                  perf_columns: List[PerfStat] = rpt.BENCHMARK_TABLE_COLUMNS,
                                  column_header: str = 'Asset',
                                  is_convert_to_str: bool = True,
                                  **kwargs
                                  ) -> pd.DataFrame:
    """
    compute ra perf table and get ra performance columns with data as string for tables
    """
    ra_perf_table = rpt.compute_ra_perf_table_with_benchmark(prices=prices,
                                                             benchmark=benchmark,
                                                             perf_params=perf_params,
                                                             **kwargs)
    df = pd.DataFrame(index=ra_perf_table.index)
    for perf_column in perf_columns:
        if is_convert_to_str:
            # here we can shorten the performance var for outputs
            df[perf_column.to_str(**kwargs)] = dfs.series_to_str(ds=ra_perf_table[perf_column.to_str()],
                                                                 var_format=perf_column.to_format(**kwargs))
        else:
            df[perf_column.to_str()] = ra_perf_table[perf_column.to_str()]

    if drop_benchmark:
        df = df.drop(benchmark, axis=0)
    df.index.name = column_header
    return df


def plot_ra_perf_table_benchmark(prices: pd.DataFrame,
                                 benchmark: str,
                                 drop_benchmark: bool = False,
                                 perf_params: PerfParams = None,
                                 perf_columns: List[PerfStat] = rpt.BENCHMARK_TABLE_COLUMNS,
                                 special_columns_colors: List[Tuple[int, str]] = ((0, 'skyblue'),),
                                 special_rows_colors: List[Tuple[int, str]] = None,
                                 column_header: str = 'Asset',
                                 fontsize: int = 10,
                                 transpose: bool = False,
                                 alpha_an_factor: float = None,
                                 is_convert_to_str: bool = True,
                                 df_to_add: pd.DataFrame = None,
                                 ax: plt.Subplot = None,
                                 **kwargs
                                 ) -> Tuple[Optional[plt.Figure], pd.DataFrame]:
    """
    plot ra perf table and get ra performance columns with data as string for tables
    """
    ra_perf_table = get_ra_perf_benchmark_columns(prices=prices,
                                                  benchmark=benchmark,
                                                  drop_benchmark=drop_benchmark,
                                                  perf_params=perf_params,
                                                  perf_columns=perf_columns,
                                                  column_header=column_header,
                                                  alpha_an_factor=alpha_an_factor,
                                                  is_convert_to_str=is_convert_to_str,
                                                  **kwargs)
    if not drop_benchmark and special_rows_colors is None:
        special_rows_colors = [(1, 'skyblue')]  # for benchmarl separation
    kwargs = sop.update_kwargs(kwargs, dict(special_rows_colors=special_rows_colors))
    
    if df_to_add is not None:
        df_to_add = df_to_add.reindex(index=ra_perf_table.index)
        if is_convert_to_str:
            df_to_add = df_to_add.fillna('')
        ra_perf_table = pd.concat([ra_perf_table, df_to_add], axis=1)

    fig = ptb.plot_df_table(df=ra_perf_table,
                            transpose=transpose,
                            special_columns_colors=special_columns_colors,
                            fontsize=fontsize,
                            ax=ax,
                            **kwargs)
    return fig, ra_perf_table


def plot_ra_perf_bars(prices: pd.DataFrame,
                      benchmark: Optional[str] = None,
                      drop_benchmark: bool = False,
                      perf_column: PerfStat = PerfStat.SHARPE_RF0,
                      perf_params: PerfParams = None,
                      legend_loc: Optional[str] = None,
                      ax: plt.Subplot = None,
                      **kwargs
                      ) -> plt.Figure:

    if benchmark is None:
        ra_perf_table = rpt.compute_ra_perf_table(prices=prices, perf_params=perf_params)
    else:
        ra_perf_table = get_ra_perf_benchmark_columns(prices=prices,
                                                      benchmark=benchmark,
                                                      drop_benchmark=drop_benchmark,
                                                      perf_params=perf_params,
                                                      is_convert_to_str=False,
                                                      **kwargs)

    df = ra_perf_table[perf_column.to_str()].to_frame()
    colors = put.compute_heatmap_colors(a=df.to_numpy())
    fig = plot_vbars(df=df,
                     var_format=perf_column.to_format(**kwargs),
                     legend_loc=legend_loc,
                     colors=colors,
                     is_category_names_colors=False,
                     ax=ax,
                     **kwargs)
    return fig


def plot_ra_perf_scatter(prices: pd.DataFrame,
                         benchmark: str = None,
                         perf_params: PerfParams = None,
                         x_var: PerfStat = PerfStat.MAX_DD,
                         y_var: PerfStat = PerfStat.PA_RETURN,
                         hue_data: pd.Series = None,
                         yvar_format: str = None,
                         x_filters: Tuple[Optional[float], Optional[float]] = None,
                         order: int = 1,
                         add_universe_model_label: bool = True,
                         ci: Optional[int] = 95,
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> plt.Figure:
    """
    scatter plot of performance stats
    """
    if benchmark is not None:
        ra_perf_table = rpt.compute_ra_perf_table_with_benchmark(prices=prices,
                                                                 benchmark=benchmark,
                                                                 perf_params=perf_params,
                                                                 **kwargs)
    else:
        ra_perf_table = rpt.compute_ra_perf_table(prices=prices, perf_params=perf_params)

    xy = ra_perf_table[[x_var.to_str(), y_var.to_str()]]
    if x_filters is not None:
        if x_filters[0] is not None:
            xy = xy.iloc[xy[x_var.to_str()].to_numpy()>x_filters[0], :]
        if x_filters[1] is not None:
            xy = xy.iloc[xy[x_var.to_str()].to_numpy()<x_filters[1], :]

    # fu.save_df_to_excel(xy, file_name='xy')
    if hue_data is not None:
        xy = pd.concat([xy, hue_data], axis=1)
        hue = hue_data.name or 'hue'
    else:
        hue = None
    fig = psc.plot_scatter(df=xy,
                           hue=hue,
                           xvar_format=x_var.to_format(**kwargs),
                           yvar_format=yvar_format or y_var.to_format(**kwargs),
                           add_universe_model_label=add_universe_model_label,
                           order=order,
                           ci=ci,
                           ax=ax,
                           **kwargs)

    return fig


def plot_ra_perf_by_dates(prices: pd.DataFrame,
                          time_period_dict: Dict[str, da.TimePeriod],
                          perf_column: PerfStat = PerfStat.SHARPE_RF0,
                          perf_params: PerfParams = None,
                          fontsize: int = 12,
                          title: str = None,
                          ax: plt.Subplot = None,
                          **kwargs
                          ) -> plt.Figure:
    """
    compute table of performance var for a dict of strat end periods
    """
    datas = []
    for key, time_period in time_period_dict.items():
        prices_ = time_period.locate(prices)
        ra_perf_table = rpt.compute_ra_perf_table(prices=prices_,
                                                  perf_params=perf_params)
        this = dfs.series_to_str(ra_perf_table[perf_column.to_str()], var_format=perf_column.to_format(**kwargs))
        datas.append(this.rename(key))
    df = pd.concat(datas, axis=1)

    fig = ptb.plot_df_table(df=df,
                            add_index_as_column=True,
                            fontsize=fontsize,
                            title=title,
                            ax=ax,
                            **kwargs)
    return fig


def plot_ra_perf_annual_matrix(price: pd.Series,
                               min_lag: int = 1,
                               date_format: str = '%d%b%Y',
                               perf_column: PerfStat = PerfStat.SHARPE_RF0,
                               perf_params: PerfParams = None,
                               fontsize: int = 12,
                               ax: plt.Subplot = None,
                               is_fig_out: bool = True,
                               is_shift_by_day: bool = False,
                               **kwargs
                               ) -> Union[plt.Figure, pd.DataFrame]:
    """
    compute annual matrix of performance var for a dict of strat end periods
    """
    if not isinstance(price, pd.Series):
        raise TypeError(f"must be pd.Series not {type(price)}")

    # extract years
    dates_schedule = da.generate_dates_schedule(time_period=da.get_time_period(df=price),
                                                freq='YE',
                                                include_start_date=True,
                                                include_end_date=True)
    yearly_dfs = {}
    for idx, start in enumerate(dates_schedule[:-min_lag]):
        yearly_df = {}
        for idxx, end in enumerate(dates_schedule[idx+min_lag:]):
            price_ = da.TimePeriod(start, end).locate(price)
            ra_perf_table = rpt.compute_ra_perf_table(prices=price_, perf_params=perf_params)
            #print(f"{start} to {end}")
            # yearly_df[end.strftime(date_format)] = f"{start} to {end}"
            if is_shift_by_day:
                this = da.shift_date_by_day(date=end, backward=False)
            else:
                this = end
            yearly_df[this.strftime(date_format)] = ra_perf_table[perf_column.to_str()].iloc[0]

        yearly_dfs[start.strftime(date_format)] = pd.Series(yearly_df)
    yearly_dfs = pd.DataFrame.from_dict(yearly_dfs, orient='index')

    if is_fig_out:
        data_colors = put.compute_heatmap_colors(a=yearly_dfs.to_numpy(), **kwargs)
        df = dfs.df_to_str(yearly_dfs, var_format=perf_column.to_format(**kwargs))
        fig = ptb.plot_df_table(df=df,
                                add_index_as_column=True,
                                index_column_name='Start \ End',
                                fontsize=fontsize,
                                data_colors=data_colors,
                                ax=ax,
                                **kwargs)
        return fig
    else:
        return yearly_dfs


def plot_desc_freq_table(df: pd.DataFrame,
                         freq: str = 'YE',
                         agg_func: Callable = np.sum,
                         var_format: str = '{:.2f}',
                         special_columns_colors: List[Tuple[int, str]] = None,
                         special_rows_colors: List[Tuple[int, str]] = None,
                         column_header: str = 'Variable',
                         fontsize: int = 12,
                         title: Optional[str] = None,
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> plt.Figure:

    data_table = rpt.compute_desc_freq_table(df=df,
                                             freq=freq,
                                             agg_func=agg_func)
    plot_data = pd.DataFrame(data=data_table.index.T, index=data_table.index, columns=[column_header])

    for column in data_table:
        plot_data[column] = data_table[column].apply(lambda x: var_format.format(x))

    fig = ptb.plot_df_table(df=plot_data,
                            special_columns_colors=special_columns_colors,
                            special_rows_colors=special_rows_colors,
                            fontsize=fontsize,
                            title=title,
                            ax=ax,
                            **kwargs)
    return fig


def plot_top_bottom_performers(prices: pd.DataFrame,
                               yvar_format: str = '{:.2%}',
                               num_assets: int = None,
                               add_bar_values: Optional[bool] = False,
                               ax: plt.Subplot = None,
                               **kwargs
                               ) -> Optional[plt.Subplot]:
    returns = ret.to_total_returns(prices=prices).sort_values().dropna()

    if num_assets is not None and len(returns.index) > 2*num_assets:
        returns = pd.concat([returns.iloc[:num_assets], returns.iloc[-num_assets:]])

    fig = plot_bars(df=returns,
                    stacked=False,
                    skip_y_axis=True,
                    legend_loc=None,
                    x_rotation=90,
                    add_bar_values=add_bar_values,
                    yvar_format=yvar_format,
                    ax=ax,
                    **kwargs)
    return fig


def plot_best_worst_returns(price: pd.Series,
                            freq: Optional[str] = None,
                            yvar_format: str = '{:.2%}',
                            date_format: str = '%d%b%Y',
                            num_returns: int = 10,
                            add_bar_values: Optional[bool] = False,
                            ax: plt.Subplot = None,
                            **kwargs
                            ) -> Optional[plt.Subplot]:
    returns = ret.to_returns(price, freq=freq).sort_values().dropna()

    if len(returns.index) > 2*num_returns:
        returns = pd.concat([returns.iloc[:num_returns], returns.iloc[-num_returns:]])

    returns.index = [t.strftime(date_format) for t in returns.index]

    fig = plot_bars(df=returns,
                    stacked=False,
                    skip_y_axis=True,
                    legend_loc=None,
                    x_rotation=90,
                    add_bar_values=add_bar_values,
                    yvar_format=yvar_format,
                    ax=ax,
                    **kwargs)
    return fig


class UnitTests(Enum):
    PLOT_RA_PERF_TABLE = 1
    PLOT_RA_PERF_SCATTER = 2
    PLOT_RA_PERF_TABLE_BENCHMARK = 3
    PLOT_DESC_FREQ_TABLE = 4
    PLOT_SHARPE_BARPLOT = 5
    PLOT_SHARPE_BY_DATES = 6
    PLOT_PERF_FOR_START_END_PERIOD = 7
    PLOT_TOP_BOTTOM_PERFORMERS = 8
    PLOT_TOP_BOTTOM_RETURNS = 9


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()
    print(prices)

    if unit_test == UnitTests.PLOT_RA_PERF_TABLE:

        perf_params = PerfParams(freq='B')
        prices = prices.iloc[:, :5]
        plot_ra_perf_table(prices=prices,
                           perf_columns=rpt.COMPACT_TABLE_COLUMNS,
                           perf_params=perf_params)

    elif unit_test == UnitTests.PLOT_RA_PERF_SCATTER:

        perf_params = PerfParams(freq='B')
        plot_ra_perf_scatter(prices=prices,
                             perf_params=perf_params)

    elif unit_test == UnitTests.PLOT_RA_PERF_TABLE_BENCHMARK:
        perf_params = PerfParams(freq='ME')
        plot_ra_perf_table_benchmark(prices=prices,
                                     benchmark='SPY',
                                     perf_params=perf_params,
                                     transpose=False)

    elif unit_test == UnitTests.PLOT_DESC_FREQ_TABLE:
        freq_data = plot_desc_freq_table(df=prices,
                                         freq='YE',
                                         agg_func=np.mean)
        print(freq_data)

    elif unit_test == UnitTests.PLOT_SHARPE_BARPLOT:
        plot_ra_perf_bars(prices=prices, perf_column=PerfStat.MAX_DD)

    elif unit_test == UnitTests.PLOT_SHARPE_BY_DATES:
        prices = prices

        time_period_dict = {'1y': da.TimePeriod(start='30Jun2019', end='30Jun2020'),
                            '3y': da.TimePeriod(start='30Jun2017', end='30Jun2020'),
                            '5y': da.TimePeriod(start='30Jun2015', end='30Jun2020')}
        plot_ra_perf_by_dates(prices=prices,
                              time_period_dict=time_period_dict)

    elif unit_test == UnitTests.PLOT_PERF_FOR_START_END_PERIOD:
        plot_ra_perf_annual_matrix(price=prices.iloc[:, 0])

    elif unit_test == UnitTests.PLOT_TOP_BOTTOM_PERFORMERS:
        plot_top_bottom_performers(prices=prices, num_assets=2)

    elif unit_test == UnitTests.PLOT_TOP_BOTTOM_RETURNS:
        plot_best_worst_returns(price=prices.iloc[:, 0])

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PLOT_RA_PERF_TABLE_BENCHMARK

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




import pandas as pd
from typing import Tuple, List
from enum import Enum
import yfinance as yf
import qis
from qis import TimePeriod, MultiPortfolioData
from qis.portfolio.reports.config import fetch_default_report_kwargs

# multiportfolio
from qis.portfolio.reports.multi_strategy_factseet_pybloqs import generate_multi_portfolio_factsheet_with_pyblogs
from qis.portfolio.reports.strategy_benchmark_factsheet_pybloqs import generate_strategy_benchmark_factsheet_with_pyblogs


def fetch_riskparity_universe_data() -> Tuple[pd.DataFrame, pd.DataFrame, pd.Series]:
    """
    define custom universe with asset class grouping
    """
    universe_data = dict(SPY='Equities',
                         QQQ='Equities',
                         EEM='Equities',
                         TLT='Bonds',
                         IEF='Bonds',
                         LQD='Credit',
                         HYG='HighYield',
                         GLD='Gold')
    tickers = list(universe_data.keys())
    group_data = pd.Series(universe_data)  # for portfolio reporting
    prices = yf.download(tickers=tickers, start=None, end=None, ignore_tz=True)['Close'][tickers]
    prices = prices.asfreq('B', method='ffill').loc['2003': ]
    benchmark_prices = prices[['SPY', 'TLT']]
    return prices, benchmark_prices, group_data


def generate_volparity_multi_strategy(prices: pd.DataFrame,
                                      benchmark_prices: pd.DataFrame,
                                      group_data: pd.Series,
                                      time_period: TimePeriod,
                                      spans: List[int] = (5, 10, 20, 40, 60, 120),
                                      vol_target: float = 0.15,
                                      rebalancing_costs: float = 0.0010
                                      ) -> MultiPortfolioData:
    """
    generate volparity sensitivity to span
    """
    returns = qis.to_returns(prices=prices, is_log_returns=True)

    portfolio_datas = []
    for span in spans:
        ra_returns, weights, ewm_vol = qis.compute_ra_returns(returns=returns, span=span, vol_target=vol_target)
        weights = weights.divide(weights.sum(1), axis=0)
        portfolio_data = qis.backtest_model_portfolio(prices=prices,
                                                      weights=time_period.locate(weights),
                                                      rebalancing_costs=rebalancing_costs,
                                                      ticker=f"VP span={span}")
        portfolio_data.set_group_data(group_data=group_data, group_order=list(group_data.unique()))
        portfolio_datas.append(portfolio_data)

    multi_portfolio_data = MultiPortfolioData(portfolio_datas, benchmark_prices=benchmark_prices)
    return multi_portfolio_data


class UnitTests(Enum):
        MULTI_PORTFOLIO = 1
        STRATEGY_BENCHMARK = 2


def run_unit_test(unit_test: UnitTests):

    time_period = qis.TimePeriod('31Dec2005', '06Sep2024')  # time period for portfolio reporting

    prices, benchmark_prices, group_data = fetch_riskparity_universe_data()
    multi_portfolio_data = generate_volparity_multi_strategy(prices=prices,
                                                             benchmark_prices=benchmark_prices,
                                                             group_data=group_data,
                                                             time_period=time_period,
                                                             vol_target=0.15,
                                                             rebalancing_costs=0.0010  # per traded volume
                                                             )

    if unit_test == UnitTests.MULTI_PORTFOLIO:

        report = generate_multi_portfolio_factsheet_with_pyblogs(multi_portfolio_data=multi_portfolio_data,
                                                  time_period=time_period,
                                                  **fetch_default_report_kwargs(time_period=time_period))

        filename = f"{qis.local_path.get_output_path()}_volparity_span_report_{pd.Timestamp.now().strftime('%Y%m%d_%H%M')}.pdf"
        report.save(filename)
        print(f"saved allocation report to {filename}")

    elif unit_test == UnitTests.STRATEGY_BENCHMARK:

        report = generate_strategy_benchmark_factsheet_with_pyblogs(multi_portfolio_data=multi_portfolio_data,
                                                                    strategy_idx=-1,
                                                                    benchmark_idx=0,
                                                                    time_period=time_period,
                                                                    **fetch_default_report_kwargs(time_period=time_period))

        filename = f"{qis.local_path.get_output_path()}_volparity_pybloq_report_{pd.Timestamp.now().strftime('%Y%m%d_%H%M')}.pdf"
        report.save(filename)
        print(f"saved allocation report to {filename}")


if __name__ == '__main__':

    unit_test = UnitTests.STRATEGY_BENCHMARK

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

import qis.plots as qp

np.random.seed(2)


def get_random_data(is_random_beta: bool = True,
                    n: int = 10000
                    ) -> pd.DataFrame:

    x = np.random.normal(0.0, 1.0, n)
    eps = np.random.normal(0.0, 1.0, n)
    if is_random_beta:
        beta = np.random.normal(1.0, 1.0, n)*np.abs(x)
    else:
        beta = np.ones(n)
    y = beta*x + eps
    df = pd.concat([pd.Series(x, name='x'), pd.Series(y, name='y')], axis=1)
    df = df.sort_values(by='x', axis=0)
    return df


df = get_random_data(n=100000)

kwargs = dict(xvar_format='{:.1f}', yvar_format='{:.1f}', order=1)

with sns.axes_style('darkgrid'):
    fig, axs = plt.subplots(1, 2, figsize=(10, 5), tight_layout=True)
    qp.plot_scatter(df=df,
                    full_sample_order=1,
                    fit_intercept=False,
                    title='Linear regression',
                    ax=axs[0],
                    **kwargs)
    qp.plot_classification_scatter(df=df,
                                   full_sample_order=None,
                                   fit_intercept=False,
                                   title='Localized sextile regression',
                                   ax=axs[1],
                                   **kwargs)
    qp.align_y_limits_ax12(ax1=axs[0], ax2=axs[1])
    qp.set_suptitle(fig, 'Estimation of noisy regression: y=beta*abs(x)*x + noise, beta=Normal(1, 1), x=Normal(0, 1), noise=Normal(0, 1)')


plt.show()


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import qis as qis
from bbg_fetch import fetch_field_timeseries_per_tickers, fetch_index_members_weights, fetch_bonds_info


index_ticker = 'I31415US Index'

end_dates = ['20180101', '20190101', '20200101', '20210101', '20220101', '20230101', '20240101', '20250101', '20250106']

data_out = {}
weighted_resets = []
for idx, end_date in enumerate(end_dates):
    if idx > 0:
        members = fetch_index_members_weights(index_ticker, END_DATE_OVERRIDE=end_dates[idx-1])
        corp_index = [f"{x} corp" for x in members.index]
        members.index = corp_index
        """
        amt_outstanding = fetch_bonds_info(isins=members.index.to_list(), fields=['amt_outstanding'],
                                           END_DATE_OVERRIDE=end_dates[idx-1])['amt_outstanding']
        amt_outstanding = amt_outstanding.loc[members.index]
        amt_outstanding.index = corp_index
        """
        prices = fetch_field_timeseries_per_tickers(tickers=corp_index,
                                                    start_date=pd.Timestamp(end_dates[idx-1]),
                                                    end_date=pd.Timestamp(end_date),
                                                    freq='B')
        prices = prices.resample('W-WED').last()
        # market_value = prices.multiply(amt_outstanding, axis=1)
        #market_value.divide(np.nansum(market_value, axis=1, keepdims=True), axis=1)
        is_reset = (prices > 100).astype(float)
        market_weights = members.iloc[:, 0]
        weighted_reset = is_reset.multiply(market_weights, axis=1)
        weighted_reset = weighted_reset.sum(1)
        weighted_resets.append(weighted_reset)
        data_out[f"{end_dates[idx-1]} members"] = members
        data_out[f"{end_date} prices"] = prices
        data_out[f"{end_date} is_reset"] = is_reset

weighted_resets = pd.concat(weighted_resets)
weighted_resets = weighted_resets.loc[~weighted_resets.index.duplicated(keep='first')]
weighted_resets = weighted_resets.to_frame('weighted_par_reset %')
print(weighted_resets)

data_out['weighted_resets'] = weighted_resets
qis.save_df_to_excel(data=data_out, file_name='perp_pricing')

with sns.axes_style("darkgrid"):
    fig, ax = plt.subplots(1, 1, figsize=(15, 8), tight_layout=True)
    qis.plot_time_series(weighted_resets,
                         title='weighted_par_reset',
                         ax=ax)

plt.show()



"""
bar plots
"""

# packages
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
from typing import List, Tuple, Optional, Union
from enum import Enum

# qis
import qis.utils.struct_ops as sop
import qis.plots.utils as put
from qis.plots.utils import LegendStats


def plot_bars(df: Union[pd.DataFrame, pd.Series],
              stacked: bool = True,
              date_format: str = '%d-%b-%y',
              x_date_freq: str = 'QE',
              title: str = None,
              fontsize: int = 10,
              add_bar_values: bool = False,
              add_top_bar_values: bool = False,
              legend_stats: LegendStats = LegendStats.NONE,
              xvar_format: str = '{:.1%}',
              yvar_format: str = '{:,.2f}',
              x_rotation: int = 90,
              skip_y_axis: bool = False,
              legend_loc: Optional[str] = 'upper center',
              bbox_to_anchor: Optional[Tuple[float, float]] = None,
              y_limits: Tuple[Optional[float], Optional[float]] = None,
              totals: List[float] = None,
              is_top_totals: bool = False,
              annotate_totals: bool = True,
              totals_offset: Tuple[float, float] = (2.55, 5),
              series_color: str = 'steelblue',
              colors: List[str] = None,
              legend_labels: List[str] = None,
              legend_colors: List[str] = None,
              vline_columns: List[int] = None,
              xlabel: str = None,
              ylabel: str = None,
              reverse_columns: bool = False,
              is_sns: bool = True,
              add_avg_line: bool = False,
              is_horizontal: bool = False,
              labels_frequency: Optional[int] = None,
              ax: plt.Subplot = None,
              **kwargs
              ) -> Optional[plt.Figure]:
    """
    plot bars
    """
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    # convert to series to avoid melting
    # if isinstance(df, pd.DataFrame) and len(df.columns) == 1:
    #    df = df.iloc[:, 0]

    if colors is None:
        if isinstance(df, pd.Series):
            colors = [series_color]
        else:
            if stacked:
                n = len(df.columns)
            else:
                n = len(df.index)
            colors = put.get_n_colors(n=n, **kwargs)

    # use str for dates with plot.bar
    if isinstance(df.index, pd.DatetimeIndex) and isinstance(df, pd.Series):
        df, datalables = put.map_dates_index_to_str(data=df,
                                                    x_date_freq=x_date_freq,
                                                    date_format=date_format)
        df.index = datalables
        if is_horizontal:
            df.plot.barh(stacked=stacked, color=colors, edgecolor='none', ax=ax)
        else:
            df.plot.bar(stacked=stacked, color=colors, edgecolor='none', ax=ax)

    elif isinstance(df, pd.Series):
        # sns.barplot(x=df.index, y=df, palette=colors, ax=ax)
        if is_horizontal:
            df.plot.barh(stacked=stacked, color=colors, edgecolor='none', ax=ax)
        else:
            df.plot.bar(stacked=stacked, color=colors, edgecolor='none', ax=ax)

    else:  # need to melt for barplot
        value_name = ylabel or 'y'
        var_name = xlabel or 'x'
        df1 = df.melt(ignore_index=False, var_name=var_name, value_name=value_name)
        df1 = df1.dropna(subset=[value_name])  # dropna in value_name
        if is_sns and not stacked:
            sns.barplot(x=df1.index, y=value_name, data=df1, hue=var_name,
                        palette=colors[:len(df1[var_name].unique())],
                        edgecolor='none',
                        orient='h' if is_horizontal else 'v',
                        ax=ax)
        else:
            if is_horizontal:
                df.plot.barh(stacked=stacked, color=colors, edgecolor='none', ax=ax)
            else:
                df.plot.bar(stacked=stacked, color=colors, edgecolor='none', ax=ax)

    # put totals to bar and store locations
    x_locs = []
    x_mins = []
    x_maxs = []
    for p in ax.patches:
        width, height = p.get_width(), p.get_height()
        y = p.get_y()
        x = p.get_x()

        x_loc = x+.2*width
        y_loc = y+.3*height if height > 0.0 else y+0.8*height
        if add_bar_values:
            if height != 0:
                ax.annotate(text=yvar_format.format(height), xy=(x_loc, y_loc), fontsize=fontsize, weight='normal')
        elif add_top_bar_values:
            if height != 0:
                ymin, ymax = ax.get_ylim()
                ax.annotate(text=yvar_format.format(height), xy=(x_loc, 0.95*ymax), fontsize=fontsize, weight='normal')

        if x not in x_locs:
            x_locs.append(x)  # take only one location per asset
            x_mins.append(x)
            x_maxs.append(x + width)

    if totals is not None:
        if is_top_totals:
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax * 1.1)

        for total, x_loc, x_min, x_max in zip(totals, x_locs, x_mins, x_maxs):
            label = yvar_format.format(total)
            if is_top_totals:
                trans = transforms.blended_transform_factory(ax.transData, ax.transAxes)
                ax.text(x_min + 0.2 * (x_max - x_min), 0.975, label,
                        transform=trans, fontsize=fontsize, weight='normal')
            else:
                ax.hlines(xmin=x_min, xmax=x_max, y=total, linestyle='-', color='black', linewidth=0.5)
                if annotate_totals:
                    ax.annotate(text=label, xytext=totals_offset, textcoords='offset points',
                                xy=(x_max, total),
                                fontsize=fontsize,
                                ha='left', va='top')

    if vline_columns is not None:
        for vline_column in vline_columns:
            ax.vlines([vline_column-0.5], *ax.get_ylim(), lw=1)  # shift by 0.5 for visibility

    if legend_labels is not None:
        labels = legend_labels
    else:
        # handles, labels = ax.get_legend_handles_labels()
        labels = put.get_legend_lines(data=df, legend_stats=legend_stats, var_format=yvar_format)

    if legend_colors is not None:
        colors = legend_colors

    put.set_legend(ax=ax,
                   labels=labels,
                   colors=colors,
                   reverse_columns=reverse_columns,
                   bbox_to_anchor=bbox_to_anchor,
                   legend_loc=legend_loc,
                   fontsize=fontsize,
                   **kwargs)

    for line in ax.get_legend().get_lines():
        line.set_linewidth(4.0)

    if add_avg_line:
        avg = np.nanmean(df)
        ax.axhline(avg, color='coral', linewidth=2, linestyle='--', label='Average')
        xmin, xmax = ax.get_xlim()
        ax.text(xmax, avg, f"Average", fontsize=fontsize, weight='normal', color='coral')

    put.set_ax_xy_labels(ax=ax,
                         fontsize=fontsize,
                         xlabel=xlabel,
                         ylabel=ylabel,
                         **kwargs)

    put.set_ax_tick_params(ax=ax)
    local_kwargs = sop.update_kwargs(dict(yvar_format=yvar_format, xvar_format=xvar_format,  fontsize=fontsize), kwargs)
    put.set_ax_ticks_format(ax=ax, x_rotation=x_rotation, **local_kwargs)
    put.set_ax_tick_labels(ax=ax, fontsize=fontsize, skip_y_axis=skip_y_axis, **kwargs)

    if is_horizontal:
        ax.set_yticks(np.arange(len(df.index)), labels=df.index.to_list())
        ax.xaxis.set_tick_params(rotation=x_rotation)
        ax.invert_yaxis()  # labels read top-to-bottom
    else:
        ax.set_xticks(np.arange(len(df.index)), labels=df.index.to_list())
        ax.xaxis.set_tick_params(rotation=x_rotation)
        ax.axhline(0, color='black', lw=0.5)

    if labels_frequency is not None:
        put.set_labels_frequency(ax=ax, labels_frequency=labels_frequency)

    if y_limits is not None:
        put.set_y_limits(ax=ax, y_limits=y_limits)

    put.set_spines(ax=ax, **kwargs)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize)

    return fig


def plot_vbars(df: Union[pd.DataFrame, pd.Series],
               title: Optional[str] = None,
               fontsize: int = 10,
               add_bar_values: bool = True,
               add_bar_perc_values: bool = False,
               var_format: str = '{:.1%}',
               legend_loc: Optional[str] = 'upper center',
               bbox_to_anchor: Optional[Tuple[float, float]] = (0.5, 1.04),
               totals: List[float] = None,
               colors: Union[List[str], List[Tuple[float, float, float]]] = None,
               legend_labels: List[str] = None,
               legend_colors: List[str] = None,
               x_rotation: int = 0,
               xmin_shift: Optional[float] = None,  # add shift to x-axis to left
               add_bar_value_at_mid: bool = True,
               add_total_bar: bool = True,
               add_total_to_index: bool = False,
               add_total_to_left: bool = False,
               is_category_names_colors: bool = True,
               x_step: Optional[float] = None,  # specify x-step
               x_limits: Tuple[Union[float, None], Union[float, None]] = None,
               reverse_columns: bool = True,
               rows_edge_lines: List[int] = None,
               ax: plt.Subplot = None,
               **kwargs
               ) -> plt.Figure:

    if isinstance(df, pd.Series):
        df = df.to_frame()
    category_names = df.columns.to_list()

    if add_total_to_index and totals is not None:
        df.index = [f"{x} {var_format.format(total)}" for x, total in zip(df.index, totals)]

    results = {rdata[0]: rdata[1].to_list() for rdata in df.iterrows()}

    labels = list(results.keys())
    np_data = np.array(list(results.values()))
    totals = np.sum(np_data, axis=1)

    if colors is None:
        if is_category_names_colors:
            if len(df.columns) == 1:
                colors = put.compute_heatmap_colors(a=df.to_numpy())
            else:
                colors = put.compute_heatmap_colors(a=np.sum(df.to_numpy(), axis=1))
        else:
            colors = put.get_n_colors(n=len(df.columns))
    else:
        legend_colors = colors
        colors = np.tile(colors, len(category_names))

    if ax is None:
        height = put.calc_table_height(num_rows=len(df.index), scale=0.30)
        fig, ax = plt.subplots(figsize=(9.2, height))
    else:
        fig = None

    if add_bar_value_at_mid:
        bar_value_at_max = None
    else:
        bar_value_at_max = np.max(np.cumsum(np_data, axis=1))

    ax.invert_yaxis()

    # negative
    last_starts = None
    initial_starts = np.sum(np.where(np_data < 0.0, np_data, 0.0), axis=1)
    for i, colname in enumerate(category_names):
        if is_category_names_colors:
            col_colors = colors[i]
        else:
            col_colors = colors
        widths = np.where(np_data[:, i] < 0.0, np_data[:, i], 0.0)
        if last_starts is None:
            starts = initial_starts
            last_starts = initial_starts
        else:
            starts = last_starts
        last_starts = last_starts + np.abs(widths)
        ax.barh(labels, np.abs(widths), left=starts, height=0.5, label=colname, color=col_colors)

        if add_bar_values:
            xcenters = starts + np.abs(widths) / 2
            text_color = 'black'
            for y, (x, c) in enumerate(zip(xcenters, widths)):
                if not np.isclose(c, 0.0):
                    if add_bar_value_at_mid:
                        x_loc = x
                    else:
                        x_loc = bar_value_at_max

                    if add_bar_perc_values:
                        label = f"{var_format.format(c)} / {'{:.0%}'.format(c/totals[y])}"
                    else:
                        label = var_format.format(c)
                    ax.text(x_loc, y, label, ha='center', va='center', color=text_color, fontsize=fontsize)

    # positive
    last_starts = 0*last_starts
    for i, colname in enumerate(category_names):
        if is_category_names_colors:
            col_colors = colors[i]
        else:
            col_colors = colors

        widths = np.where(np_data[:, i] > 0.0, np_data[:, i], 0.0)
        if last_starts is None:
            starts = 0
            last_starts = widths
        else:
            starts = last_starts
            last_starts = last_starts + widths
        ax.barh(labels, widths, left=starts, height=0.5, label=colname, color=col_colors)

        if add_bar_values:
            xcenters = starts + widths / 2
            text_color = 'black'
            for y, (x, c) in enumerate(zip(xcenters, widths)):
                if not np.isclose(c, 0.0):
                    if add_bar_value_at_mid:
                        x_loc = x
                    else:
                        x_loc = bar_value_at_max
                    if add_bar_perc_values:
                        label = f"{var_format.format(c)} / {'{:.0%}'.format(c/totals[y])}"
                    else:
                        label = var_format.format(c)

                    ax.text(x_loc, y, label, ha='center', va='center', color=text_color, fontsize=fontsize)

    if xmin_shift is not None:
        xmin, xmax = ax.get_xlim()
        xmin_ = xmin + xmin_shift
        ax.set_xlim(xmin_, xmax)

    if add_total_bar:
        for idx, total in enumerate(totals):
            ax.vlines(x=total, ymin=idx-0.25, ymax=idx+0.25, linestyle='-', color='black', linewidth=2)

    if add_total_to_left:
        widths = np.nansum(np.where(np_data > 0.0, np_data, 0.0), axis=1)
        shift = np.maximum(0.2 * np.max(widths), 0.2)
        for idx, total in enumerate(totals):
            label = f"total: {var_format.format(total)}"
            ax.text(widths[idx]+shift, idx, label, ha='center', va='center', fontsize=fontsize)

    # legend
    if legend_labels is None:
        legend_labels = category_names
        # handles, labels = ax.get_legend_handles_labels()

    if legend_colors is not None:
        legend_colors = legend_colors
    else:
        legend_colors = colors

    # reverse
    if reverse_columns:
        legend_labels = legend_labels[::-1]
        legend_colors = legend_colors[::-1]

    put.set_legend(ax=ax,
                   labels=legend_labels,
                   colors=legend_colors,
                   legend_loc=legend_loc,
                   reverse_columns=True,
                   bbox_to_anchor=bbox_to_anchor,
                   fontsize=fontsize,
                   **kwargs)
    # increase line width
    for line in ax.get_legend().get_lines():
        line.set_linewidth(5.0)

    if x_limits is not None:
        put.set_x_limits(ax=ax, x_limits=x_limits)

    # ax.xaxis.set_visible(False)
    ax.grid(zorder=0, axis='x')
    ax.axvline(x=0, linewidth=2, color='orange')

    x_labels = [var_format.format(x) for x in ax.get_xticks()]
    put.set_ax_tick_labels(ax=ax,
                           x_labels=x_labels,
                           y_labels=labels,
                           fontsize=fontsize,
                           x_rotation=x_rotation,
                           **kwargs)

    put.set_spines(ax=ax, **kwargs)

    if x_step is not None:
        x_limits = (x_step*np.floor(np.min(np.cumsum(np.where(np_data < 0.0, np_data, 0.0), axis=1))/x_step),
                    x_step*np.ceil(np.max(np.cumsum(np.where(np_data > 0.0, np_data, 0.0), axis=1))/x_step))

        put.set_x_limits(ax=ax, x_limits=x_limits)

    if rows_edge_lines is not None:
        for rows_edge_line in rows_edge_lines:
            ax.axhline(y=rows_edge_line-0.5, color='black', alpha=0.5)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize, **kwargs)

    return fig


class UnitTests(Enum):
    BARS = 1
    BARS2 = 2
    TOP_BOTTOM_RETURNS = 3
    VBAR_WEIGHTS = 4
    MONTHLY_RETURNS_BARS = 5


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.BARS:
        n = 11
        index = [f"id{x+1} {x**2}" for x in range(n)]
        df1 = pd.DataFrame(np.linspace(0.0, 1.0, n), index=index, columns=['part1'])
        df2 = pd.DataFrame(np.linspace(0.0, 0.5, n), index=index, columns=['part2'])
        df = pd.concat([df1, df2], axis=1)
        plot_bars(df=df, stacked=True)

    elif unit_test == UnitTests.BARS2:

        n = 11
        index = [f"id{x+1} {x**2}" for x in range(n)]
        df1 = pd.DataFrame(np.linspace(0.0, 1.0, n), index=index, columns=['data1'])
        df2 = pd.DataFrame(np.linspace(0.0, 0.5, n), index=index, columns=['data2'])

        fig, axs = plt.subplots(1, 2, figsize=(8, 6), tight_layout=True)
        datas = [df1, df2]
        titles = ['Group Lasso', 'Lasso']
        for data, ax, title in zip(datas, axs, titles):
            plot_bars(df=data,
                      stacked=False,
                      skip_y_axis=True,
                      title=title,
                      legend_loc=None,
                      x_rotation=90,
                      ax=ax)
        put.align_y_limits_ax12(ax1=axs[0], ax2=axs[1], is_invisible_y_ax2=True)

    elif unit_test == UnitTests.TOP_BOTTOM_RETURNS:

        from qis.test_data import load_etf_data
        import qis.perfstats.returns as ret

        prices = load_etf_data().dropna().loc['2021', :]
        returns = ret.to_total_returns(prices=prices).sort_values()
        print(returns)
        fig, ax = plt.subplots(1, 1, figsize=(8, 6), tight_layout=True)

        plot_bars(df=returns,
                  stacked=False,
                  skip_y_axis=True,
                  legend_loc=None,
                  x_rotation=90,
                  ax=ax)

    elif unit_test == UnitTests.VBAR_WEIGHTS:
        desc_dict = {'f1': (0.5, 0.5),
                     'f2': (0.7, 0.3),
                     'f3': (1.0, 0.0),
                     'f4': (0.8, 0.2),
                     'f5': (0.35, 0.65),
                     'f6': (0.0, 1.0)}

        df = pd.DataFrame.from_dict(desc_dict, orient='index', columns=['as1', 'as2'])
        print(df)
        plot_vbars(df=df,
                   colors=put.get_n_colors(n=len(df.columns)),
                   bbox_to_anchor=(0.5, 1.25),
                   add_bar_values=False,
                   add_bar_value_at_mid=False,
                   add_total_bar=False)

    elif unit_test == UnitTests.MONTHLY_RETURNS_BARS:
        from qis.test_data import load_etf_data
        import qis.perfstats.returns as ret

        prices = load_etf_data().dropna().loc['2020':, :].iloc[:, :3]
        returns = ret.to_returns(prices=prices, freq='ME', drop_first=True)
        print(returns)
        fig, ax = plt.subplots(1, 1, figsize=(8, 6), tight_layout=True)

        plot_bars(df=returns,
                  stacked=False,
                  skip_y_axis=True,
                  x_rotation=90,
                  yvar_format='{:,.0%}',
                  date_format='%b-%y',
                  fontsize=6,
                  ax=ax)
    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.VBAR_WEIGHTS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
backtest example of leveraged strategies
"""

# packages
import pandas as pd
import matplotlib.pyplot as plt
import yfinance as yf
import qis as qis

# select tickers
benchmark = 'SPY'
tickers = [benchmark, 'SSO', 'IEF']

# fetch prices
prices = yf.download(tickers=tickers, start=None, end=None, ignore_tz=True)['Close'][tickers]
prices = prices.asfreq('B', method='ffill').dropna()  # make B frequency

rebalancing_freq = 'B'  # each business day
rebalancing_costs = 0.0010  # 10bp for rebalancing

# 50/50 SSO/IEF
unleveraged_portfolio = qis.backtest_model_portfolio(prices=prices[['SSO', 'IEF']],
                                                     weights={'SSO': 0.5, 'IEF': 0.5},
                                                     rebalancing_freq=rebalancing_freq,
                                                     rebalancing_costs=rebalancing_costs,
                                                     ticker='50/50 SSO/IEF').get_portfolio_nav()

# leveraged is funded at 100bp + 3m UST
funding_rate = 0.01 + yf.download('^IRX', start=None, end=None)['Close'].dropna() / 100.0
leveraged_portfolio = qis.backtest_model_portfolio(prices=prices[['SPY', 'IEF']],
                                                   weights={'SPY': 1.0, 'IEF': 0.5},
                                                   rebalancing_freq=rebalancing_freq,
                                                   rebalancing_costs=rebalancing_costs,
                                                   funding_rate=funding_rate,
                                                   ticker='100/50 SPY/IEF').get_portfolio_nav()


prices = pd.concat([prices, unleveraged_portfolio, leveraged_portfolio], axis=1)

# generate report since SSO launch
time_period = qis.TimePeriod('21Jun2006', '01Sep2023')
fig = qis.generate_multi_asset_factsheet(prices=prices,
                                         benchmark=benchmark,
                                         time_period=time_period,
                                         **qis.fetch_default_report_kwargs(time_period=time_period))

qis.save_fig(fig=fig, file_name=f"leveraged_fund_analysis", local_path=qis.local_path.get_output_path())

qis.save_figs_to_pdf(figs=[fig],
                     file_name=f"leveraged_fund_analysis", orientation='landscape',
                     local_path=qis.local_path.get_output_path())

plt.show()




# packages
import numpy as np
import pandas as pd
from enum import Enum
from typing import Optional
import qis


class OhlcEstimatorType(Enum):
    PARKINSON = 'Parkinson'
    GARMAN_KLASS = 'Garman-Klass'
    ROGERS_SATCHELL = 'Rogers-Satchell'
    CLOSE_TO_CLOSE = 'Close-to-Close'


def estimate_ohlc_var(ohlc_data: pd.DataFrame,  # must contain ohlc columnes
                      ohlc_estimator_type: OhlcEstimatorType = OhlcEstimatorType.PARKINSON,
                      min_size: int = 2
                      ) -> pd.Series:

    if ohlc_data.empty or len(ohlc_data.index) < min_size:
        return np.nan

    log_ohlc = np.log(ohlc_data[['open', 'high', 'low', 'close']].to_numpy())
    open, high, low, close = log_ohlc[:, 0], log_ohlc[:, 1], log_ohlc[:, 2], log_ohlc[:, 3]

    hc = high - close
    ho = high - open
    lc = low - close
    lo = low - open
    hl = high - low
    co = close - open

    if ohlc_estimator_type == OhlcEstimatorType.CLOSE_TO_CLOSE:
        sample_var = np.square(close[1:]-close[:-1])
        sample_var = np.concatenate((np.array([np.nan]), sample_var), axis=0)

    elif ohlc_estimator_type == OhlcEstimatorType.PARKINSON:
        multiplier = 1.0 / (4.0 * np.log(2.0))
        sample_var = multiplier * np.square(hl)

    elif ohlc_estimator_type == OhlcEstimatorType.GARMAN_KLASS:
        multiplier = 2.0 * np.log(2.0) - 1.0
        sample_var = 0.5 * np.square(hl) - multiplier * np.square(co)

    elif ohlc_estimator_type == OhlcEstimatorType.ROGERS_SATCHELL:
        sample_var = hc*ho + lc*lo

    else:
        raise TypeError(f"unknown ohlc_estimator_type={ohlc_estimator_type}")

    sample_var = pd.Series(sample_var, index=ohlc_data.index)
    return sample_var


def estimate_hf_ohlc_vol(ohlc_data: pd.DataFrame,
                         ohlc_estimator_type: OhlcEstimatorType = OhlcEstimatorType.PARKINSON,
                         annualization_factor: float = None,  # annualisation factor highly recomended
                         is_exclude_weekends: bool = False,  # for crypto
                         agg_freq: Optional[str] = 'B'
                         ) -> pd.Series:
    """

    group hf data into daily or higher frequency bins
    for each sample compute vol at data freq and annualize at an
    """
    sample_var = estimate_ohlc_var(ohlc_data=ohlc_data, ohlc_estimator_type=ohlc_estimator_type)
    if agg_freq is not None:
        sample_var = sample_var.resample(agg_freq).mean()

    if annualization_factor is None:
        annualization_factor = qis.infer_an_from_data(data=sample_var)

    vols = np.sqrt(annualization_factor*sample_var)
    if is_exclude_weekends:
        vols = vols[vols.index.dayofweek < 5]
    return vols



# packages
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import yfinance as yf
from enum import Enum
from typing import List
import qis as qis
from qis import PerfStat


# define entries to show in ra perf table

RA_TABLE_COLUMNS = (PerfStat.START_DATE,
                    PerfStat.END_DATE,
                    PerfStat.TOTAL_RETURN,
                    PerfStat.PA_RETURN,
                    PerfStat.VOL,
                    PerfStat.SHARPE_RF0,
                    PerfStat.SHARPE_EXCESS,
                    PerfStat.MAX_DD,
                    PerfStat.MAX_DD_VOL,
                    PerfStat.SKEWNESS,
                    PerfStat.KURTOSIS,
                    PerfStat.ALPHA,
                    PerfStat.BETA,
                    PerfStat.R2)


def generate_performances(prices: pd.DataFrame,
                          regime_benchmark: str,
                          perf_params: qis.PerfParams = None,
                          perf_columns: List[PerfStat] = RA_TABLE_COLUMNS,
                          heatmap_freq: str = 'YE',
                          **kwargs
                          ) -> None:

    local_kwargs = dict(digits_to_show=1,
                        framealpha=0.75,
                        perf_stats_labels=qis.PerfStatsLabels.DETAILED_WITH_DD.value)
    kwargs = qis.update_kwargs(kwargs, local_kwargs)

    fig, ax = plt.subplots(1, 1, figsize=(12, 4), tight_layout=True)

    qis.plot_ra_perf_table_benchmark(prices=prices,
                                     benchmark=regime_benchmark,
                                     perf_params=perf_params,
                                     perf_columns=perf_columns,
                                     title=f"Risk-adjusted performance: {qis.get_time_period_label(prices, date_separator='-')}",
                                     ax=ax,
                                     **kwargs)

    fig, ax = plt.subplots(1, 1, figsize=(7, qis.calc_table_height(num_rows=len(prices.columns)+5, scale=0.5)), tight_layout=True)
    qis.plot_periodic_returns_table(prices=prices,
                                    freq=heatmap_freq,
                                    ax=ax,
                                    title=f"Periodic performance: {qis.get_time_period_label(prices, date_separator='-')}",
                                    total_name='Total',
                                    **qis.update_kwargs(kwargs, dict(square=False, x_rotation=90)))

    with sns.axes_style("darkgrid"):
        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        qis.plot_prices(prices=prices,
                        regime_benchmark=regime_benchmark,
                        perf_params=perf_params,
                        title=f"Time series of $1 invested: {qis.get_time_period_label(prices, date_separator='-')}",
                        ax=ax,
                        **kwargs)

        fig, axs = plt.subplots(2, 1, figsize=(7, 7))
        qis.plot_prices_with_dd(prices=prices,
                                regime_benchmark=regime_benchmark,
                                perf_params=perf_params,
                                title=f"Time series of $1 invested: {qis.get_time_period_label(prices, date_separator='-')}",
                                axs=axs,
                                **kwargs)

        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        qis.plot_scatter_regression(prices=prices,
                                    regime_benchmark=regime_benchmark,
                                    regime_params=qis.BenchmarkReturnsQuantileRegimeSpecs(freq=perf_params.freq_reg),
                                    perf_params=perf_params,
                                    title=f"Regime Conditional Regression: {qis.get_time_period_label(prices, date_separator='-')}",
                                    ax=ax,
                                    **kwargs)


class UnitTests(Enum):
    ETF_DATA = 1
    CRYPTO_DATA = 2
    TF_ETF = 3
    ETFS = 4
    COMMODITY_ETFS = 5
    VOL_ETFS = 6


def run_unit_test(unit_test: UnitTests):

    ust_3m_rate = yf.download('^IRX', start=None, end=None)['Close'].dropna() / 100.0

    if unit_test == UnitTests.ETF_DATA:
        regime_benchmark = 'SPY'
        tickers = [regime_benchmark, 'QQQ', 'EEM', 'TLT', 'IEF', 'LQD', 'HYG', 'SHY', 'GLD']

    elif unit_test == UnitTests.CRYPTO_DATA:
        regime_benchmark = 'BTC-USD'
        tickers = [regime_benchmark, 'ETH-USD', 'SOL-USD']
        regime_benchmark = 'BTC-USD'
        tickers = [regime_benchmark, 'SPY', 'TLT', 'ETH-USD', 'SOL-USD']

    elif unit_test == UnitTests.TF_ETF:
        regime_benchmark = 'SPY'
        tickers = [regime_benchmark, 'DBMF', 'WTMF', 'CTA']

    elif unit_test == UnitTests.ETFS:
        regime_benchmark = 'AOR'
        tickers = [regime_benchmark, 'SPY', 'PEX', 'PSP', 'GSG', 'COMT', 'REET', 'REZ']

    elif unit_test == UnitTests.COMMODITY_ETFS:
        regime_benchmark = 'AOR'
        tickers = [regime_benchmark, 'SPY', 'GLD', 'GSG', 'COMT', 'PDBC']

    elif unit_test == UnitTests.VOL_ETFS:
        regime_benchmark = 'SPY'
        tickers = [regime_benchmark, 'SVOL']

    else:
        raise NotImplementedError

    is_long_period = True
    if is_long_period:
        time_period = None
        time_period = qis.TimePeriod('31Dec2015', None)
        time_period = qis.TimePeriod('16Oct2014', None)
        # time_period = qis.TimePeriod('31Dec2017', '31Mar2023')
        perf_params = qis.PerfParams(freq='W-WED', freq_reg='ME', freq_drawdown='B', rates_data=ust_3m_rate, alpha_an_factor=12)
        kwargs = dict(x_date_freq='YE', heatmap_freq='YE', date_format='%Y', perf_params=perf_params)
    else:
        time_period = qis.TimePeriod('31Dec2022', None)
        perf_params = qis.PerfParams(freq='W-WED', freq_reg='W-WED', freq_drawdown='B', rates_data=ust_3m_rate, alpha_an_factor=52)
        kwargs = dict(x_date_freq='ME', heatmap_freq='ME', date_format='%b-%y', perf_params=perf_params)

    prices = yf.download(tickers, start=None, end=None)['Close'][tickers].dropna()

    if time_period is not None:
        prices = time_period.locate(prices)

    generate_performances(prices=prices,
                          regime_benchmark=regime_benchmark,
                          **kwargs)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.VOL_ETFS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)
